#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Logging Handler

This module initializes the logging intercept handler, which intercepts other logging methods and translates them into
 the intended loguru logging format.
"""

import logging
import sys
from pathlib import Path
from tempfile import gettempdir

from loguru import logger

from weather_provider_api.config import APP_CONFIG, APP_DEBUGGING, APP_LOG_LEVEL


class LoggingInterceptHandler(logging.Handler):
    """A class with the purpose of intercepting all log messages, and forwarding it to our own custom logging system.

    Notes:
        For our custom logging, we use Loguru.

    """

    log_level_map = {  # A level-translation map to translate numerical logging levels to string-based levels.
        50: "CRITICAL",
        40: "ERROR",
        30: "WARNING",
        20: "INFO",
        10: "DEBUG",
        0: "NOTSET",
    }

    def emit(self, record: logging.LogRecord) -> None:
        """Emission method that intercepts the original logging message using Python's base logging system and  # noqa: D205
         re-formats it (if needed) to match Loguru systems.

        Args:
            record: The original logging LogRecord object that holds a record to be converted.

        Returns:
            Nothing. We bind the converted message directly to de Loguru logger for processing.

        """
        # Set the proper level by name if possible, otherwise by using the translation map 'log_level_map'
        try:
            level = logger.level(record.levelname).name
        except AttributeError:
            level = self.log_level_map[record.levelno]

        # Set the proper log depth and frame to use
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        # Log the message as intended by writing the converted data to the proper log area.
        log = logger.bind(request_id="app")
        log.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def initialize_logging():
    """The method that initializes and sets our custom logging system as the default, and reroutes other logging  # noqa: D205
     systems to use our system instead.

    Returns:
        Nothing. The logging system itself is overwritten.
    """
    logger.remove()  # Remove any existing default loggers from Loguru

    # Log formats:
    #  - [Elastic]: Holds the official format set as the Chapter standard by Data & Analytics.
    #  - [Develop]: Holds a more elaborate and colorful version of the log, good for debugging and development.
    log_format_elastic = "{level} [UTC {time:YYYY-MM-DD HH:mm:ss:SSS!UTC}] {message}"
    log_format_develop = (
        "<level>[{level: <8}]</level>"
        "<green>[{time:YYYY-MM-DD HH:mm:ss:SSS!UTC} UTC]</green> "
        "<cyan><i>{name}</i></cyan>|<blue><b>{function}</b></blue>: <level>{message}</level>"
    )

    log_format = log_format_develop if APP_DEBUGGING else log_format_elastic

    # Hook up the stdout logger if configured
    if APP_CONFIG["logging"]["use_stdout_logger"]:
        logger.add(
            sys.stdout,
            enqueue=False,
            backtrace=False,
            level=APP_LOG_LEVEL,
            format=log_format,
            diagnose=False,
        )

    # Hook up the file loggers if configured
    if APP_CONFIG["logging"]["use_file_loggers"]:
        # Determine log folder to use
        if Path(APP_CONFIG["logging"]["file_logger_folder"]).parent.exists():
            log_folder = Path(APP_CONFIG["logging"]["file_logger_folder"])
        else:
            log_folder = Path(gettempdir()).joinpath("logs")
        log_folder.mkdir(exist_ok=True)

        # Debug
        logger.add(
            log_folder.joinpath("debug.log"),
            enqueue=True,
            filter=lambda record: record["level"].name == "DEBUG",
            format=log_format,
        )
        # Info
        logger.add(
            log_folder.joinpath("info.log"),
            enqueue=True,
            filter=lambda record: record["level"].name == "INFO",
            format=log_format,
        )
        # Error
        logger.add(
            log_folder.joinpath("error.log"),
            enqueue=True,
            filter=lambda record: record["level"].name == "ERROR",
            format=log_format,
        )
        # Regular log
        logger.add(
            log_folder.joinpath("memo.log"),
            level=APP_LOG_LEVEL,
            format=log_format,
        )

    # Because Uvicorn already uses its own logger, we'll also need to replace those:
    existing_uvicorn_loggers = (
        logging.getLogger(name) for name in logging.root.manager.loggerDict if name.startswith("uvicorn.")
    )
    for uvicorn_logger in existing_uvicorn_loggers:
        uvicorn_logger.handlers = [LoggingInterceptHandler()]

    # Finally, we hook up the LoggingInterceptHandler as the default handler
    logging.basicConfig(handlers=[LoggingInterceptHandler()], level=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""KNMI hour models data fetcher."""

import copy
import json
from datetime import datetime
from typing import List, Optional

import numpy as np
import pandas as pd
import requests
import xarray as xr
from dateutil.relativedelta import relativedelta
from loguru import logger
from pytz import UTC

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.stations import stations_history
from weather_provider_api.routers.weather.sources.knmi.utils import (
    find_closest_stn_list,
)
from weather_provider_api.routers.weather.utils.date_helpers import (
    validate_begin_and_end,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.pandas_helpers import coords_to_pd_index


class UurgegevensModel(WeatherModelBase):
    """A Weather Model that incorporates the: KNMI Uurgegevens dataset into the Weather Provider API."""

    def __init__(self):
        super().__init__()
        self.id = "uurgegevens"
        self.name = "KNMI uurgegevens"
        self.version = ""
        self.url = "https://daggegevens.knmi.nl/klimatologie/uurgegevens"
        self.predictive = False
        self.time_step_size_minutes = 1440
        self.num_time_steps = 0
        self.description = (
            "Hourly weather measurements. Can be returned for a specified period. "
            "The number of measurements returned depends on this period selection."
        )
        self.async_model = False
        self.download_url = "https://daggegevens.knmi.nl/klimatologie/uurgegevens"

        self.to_si = {
            "DD": {"convert": self.no_conversion},
            "DDVEC": {
                "name": "wind_direction",
                "convert": self.no_conversion,
            },  # 360 N, 270 W ...
            "FH": {
                "name": "wind_speed",
                "convert": self.normalize_tenths,
            },  # 0.1 m/s -> m/s
            "FF": {"convert": self.normalize_tenths},  # 0.1 m/s -> m/s
            "FX": {
                "name": "wind_speed_max",
                "convert": self.normalize_tenths,
            },  # 0.1 m/s -> m/s
            "T": {
                "name": "temperature",
                "convert": self.tenth_celsius_to_kelvin,
            },  # 0.1 degree C -> Kelvin
            "T10N": {
                "name": "temperature_min_10cm",
                "convert": self.tenth_celsius_to_kelvin,
            },  # 0.1 degree C -> Kelvin
            "TD": {"convert": self.tenth_celsius_to_kelvin},  # 0.1 degree C -> Kelvin
            "SQ": {
                "name": "sunlight_duration",
                "convert": self.normalize_tenths,
            },  # 0.1 hour -> hour
            "Q": {
                "name": "global_radiation",
                "convert": lambda x: x * 1e4,
            },  # J/cm**2 -> J/m**2
            "DR": {
                "name": "precipitation_duration",
                "convert": self.normalize_tenths,
            },  # 0.1 hour -> hour
            "RH": {
                "name": "precipitation",
                "convert": lambda x: x / 10 / 1000,
            },  # 0.1 mm -> m
            "P": {
                "name": "air_pressure",
                "convert": lambda x: x / 10 * 100,
            },  # 0.1 hPa -> Pa
            "VV": {
                "name": "visibility",
                "convert": self.knmi_visibility_class_to_meter_estimate,
            },
            "N": {
                "name": "cloud_cover",
                "convert": self.no_conversion,
            },  # orig[1,2...0] -> orig[1,2...0]
            "U": {
                "name": "humidity",
                "convert": self.percentage_to_frac,
            },  # % -> fraction
            "WW": {"convert": self.no_conversion},
            "IX": {"convert": self.no_conversion},
            "M": {"name": "fog_occurred", "convert": self.no_conversion},
            "R": {"name": "rain_occurred", "convert": self.no_conversion},
            "S": {"name": "snow_occurred", "convert": self.no_conversion},
            "O": {"name": "lightning_occurred", "convert": self.no_conversion},
            "Y": {"name": "icing_occurred", "convert": self.no_conversion},
        }

        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["T"]["convert"] = self.normalize_tenths  # 0.1 degree C -> C
        self.to_human["T10N"]["convert"] = self.normalize_tenths  # 0.1 degree C -> C
        self.to_human["TD"]["convert"] = self.normalize_tenths  # 0.1 degree C -> C
        self.to_human["RH"]["convert"] = self.normalize_tenths  # 0.1 mm -> mm

        self.human_to_model_specific = self._create_reverse_lookup(self.to_si)

        self.knmi_aliases = [
            "WIND",
            "TEMP",
            "SUNR",
            "PRCP",
            "PRES",
            "VICL",
            "MSTR",
            "ALL",
        ]

        logger.debug(f"Weather model [{self.id}] initialized successfully")

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: datetime,
        end: datetime,
        weather_factors: List[str] | None = None,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested Daggegevens weather data from the KNMI site
        and returns it as a Xarray Dataset.

        (Though this model downloads from a specific download url, the question remains whether this source is also
        listed on the new KNMI Data Platform).

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.
        """
        # Test and account for invalid datetime timeframes or input
        begin, end = validate_begin_and_end(begin, end, None, datetime.utcnow() - relativedelta(days=1))
        # Get a list of the relevant STNs and choose the closest STN for each coordinate
        station_id, stns, _ = find_closest_stn_list(stations_history, coords)

        # Download the weather data for the relevant STNs
        raw_data = self._download_weather(
            stations=stns,
            start=begin,
            end=end,
            weather_factors=weather_factors,
        )

        # Parse the raw data into a Dataset
        raw_ds = self._parse_raw_weather_data(raw_data)

        # Prepare and format the weather data for output
        ds = self._prepare_weather_data(coords, station_id, raw_ds)

        # The KNMI model isn't working properly yet, so we have to cut out any overflow time-wise
        ds = ds.sel(time=slice(begin.astimezone(UTC).replace(tzinfo=None), end.astimezone(UTC).replace(tzinfo=None)))
        return ds

    def is_async(self):  # pragma: no cover
        return self.async_model

    def _download_weather(
        self,
        stations: List[int],
        start: datetime,
        end: datetime,
        weather_factors: List[str] | None = None,
    ):
        """A function that downloads the weather from the KNMI download location and returns it as a text
        Args:
            stations:           A list containing the requested stations
            start:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            A field containing the full response of the made download-request (text-based)
        """
        # fetch data
        params = self._create_request_params(start, end, stations, weather_factors)
        r = requests.post(url=self.download_url, data=params, timeout=60)

        if r.status_code != 200:
            raise requests.HTTPError(
                "Failed to retrieve data from the KNMI website",
                r.status_code,
                self.download_url,
                params,
            )
        elif r.text == "[]":
            raise ValueError(
                "No data was returned for this request. Make that the requested data exist for this "
                "dataset, and that is was properly requested. In case of doubt, contact support."
            )
        return r.text

    def _create_request_params(
        self,
        start: datetime,
        end: datetime,
        stations: list[int],
        weather_factors: list[str],
    ):
        """A Function that transforms the request settings into parameters usable for the KMNI download request.

        Args:
            start:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            stations:           A list containing the requested stations
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            A params field (string) containing the matching settings for a KNMI Daggegevens download request.
        """
        params = {
            "fmt": "json",
            "stns": ":".join(str(station) for station in stations),
            "start": f"{start.strftime('%Y%m%d')}01",
            "end": f"{end.strftime('%Y%m%d')}24",
        }
        if weather_factors is None:
            weather_factors = ["ALL"]

        updated_weather_factors = self._request_weather_factors(weather_factors)

        params["vars"] = ":".join(updated_weather_factors)

        return params

    def _parse_raw_weather_data(self, raw_data: str) -> xr.Dataset:
        json_data = json.loads(raw_data)
        dataframe_data = pd.DataFrame.from_dict(json_data, orient="columns")

        conversion_dict = {
            "hour": str,
            "station_code": int,
        }
        for weather_factor in self.to_si.keys():
            if weather_factor in dataframe_data.keys():
                conversion_dict[weather_factor] = np.float64

        # KNMI measures the -th hour. (The 24th hour is from 23:00 to 00:00 the next day) We use 23:00 to indicate that.
        dataframe_data["hour"] = dataframe_data["hour"] - 1
        dataframe_data = dataframe_data.astype(conversion_dict)
        dataframe_data["date"] = pd.to_datetime(dataframe_data["date"])

        # Convert hours from time to timestamp
        dataframe_data["timestamp"] = pd.to_timedelta(dataframe_data["hour"] + ":00:00")

        # Merge the hours with the date field and drop the timestamp and hour fields
        dataframe_data["date"] = dataframe_data["date"] + dataframe_data["timestamp"]
        dataframe_data = dataframe_data.drop(["hour", "timestamp"], axis=1)

        dataframe_data = dataframe_data.set_index(["station_code", "date"])

        return dataframe_data.to_xarray()

    @staticmethod
    def _prepare_weather_data(coordinates: List[GeoPosition], station_id: list[np.int64], raw_ds: xr.Dataset):
        """A function that prepares the weather data for return by the API, by replacing the matching station with the
        lat/lon location that was requested, and properly formatting the dimensions.

        """
        # re-arrange stns
        ds = raw_ds.sel(station_code=station_id)

        # dict of data
        data_dict = {var_name: (["coord", "time"], var.values) for var_name, var in ds.data_vars.items()}
        timeline = pd.DatetimeIndex(ds.coords["date"].values)

        ds = xr.Dataset(
            data_vars=data_dict,
            coords={"time": timeline, "coord": coords_to_pd_index(coordinates)},
        )
        ds = ds.unstack("coord")
        return ds

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        """Implementation of the Base Weather Model function that returns a list of known weather factors for the
        model.
        """
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_up = f.upper()
            f_low = f.lower()

            if f_up in self.to_si:
                # KNMI daggegevens and uurgegevens (TD, TG, FHX, ...)
                new_factors.append(f_up)
            elif f_low in self.human_to_model_specific:
                # KNMI daggegevens and uurgegevens with human names
                new_factors.append(self.human_to_model_specific[f_low])
            else:
                try:
                    if f_up in self.knmi_aliases:
                        new_factors.append(f_up)
                except AttributeError:
                    continue

        return list(set(new_factors))  # Cleanup any duplicate values and return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Prometheus Middleware handler"""

from fastapi import FastAPI
from loguru import logger
from starlette_prometheus import PrometheusMiddleware, metrics


def initialize_prometheus_interface(application: FastAPI, metrics_endpoint: str = "/metrics"):
    """The method that attaches the Prometheus Middleware to a FastAPI application.

    Args:
        application:        The FastAPI application to attach the Prometheus Middleware to.
        metrics_endpoint:   The endpoint at which the Prometheus data should become available.

    Returns:
        Nothing. The FastAPI application itself is updated.

    Notes:
        Prometheus is a metrics system that interprets requests and results and processes those into metrics. From a
         Prometheus server this data can be gathered and preprocessed, allowing for more extensive data on request
         and the handling thereof, like information on memory usage or the time taken to process certain types of
         requests.
        Note that the Prometheus server itself also doesn't handle processing that output into views or graphs. To
         visualize Prometheus output, you'd need something like Elastic or Grafana.
    """
    application.add_middleware(PrometheusMiddleware, filter_unhandled_paths=True)
    application.add_route(metrics_endpoint, metrics)
    logger.info(
        "Attached Prometheus Middleware to the application and opened metrics endpoint at " f"[{metrics_endpoint}]..."
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

# factor name mapping

import json

from weather_provider_api.routers.weather.utils.file_helpers import (
    get_var_map_file_location,
)

file_to_use = get_var_map_file_location("arome_var_map.json")

with open(file_to_use, "r") as _f:
    arome_factors: dict = json.load(_f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from typing import List, Tuple

import numpy as np

from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


def round_coordinates_to_wgs84_grid(
    coordinates: List[GeoPosition],
    grid_resolution_lat_lon: Tuple[float, float],
    starting_points_lat_lon: Tuple[float, float] = (0, 0),
) -> List[GeoPosition]:
    """A function that rounds coordinates to a WGS84 coordinate grid based on the given settings.

    Args:
        coordinates:                The list of coordinates to round to the grid.
        grid_resolution_lat_lon:    A tuple holding the grid resolution to use for rounding
        starting_points_lat_lon:    An optional tuple holding the starting values of the grid.
                                     Useful if a grid doesn't start at a multitude of the grid resolution.

    Returns:
        A list of coordinates that have been rounded to the nearest points on the given grid.

    """
    grid_res_lat = grid_resolution_lat_lon[0]
    grid_res_lon = grid_resolution_lat_lon[1]
    start_lat = starting_points_lat_lon[0]
    start_lon = starting_points_lat_lon[1]

    wgs84_coordinate_list = [coordinate.get_WGS84() for coordinate in coordinates]
    rounded_wgs84_coordinate_list = [
        (
            (np.round((coordinate[0] - start_lat) / grid_res_lat) * grid_res_lat) + start_lat,
            (np.round((coordinate[1] - start_lon) / grid_res_lon) * grid_res_lon) + start_lon,
        )
        for coordinate in wgs84_coordinate_list
    ]

    return [GeoPosition(coordinate[0], coordinate[1]) for coordinate in rounded_wgs84_coordinate_list]



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def prepare_example_response(example_response, content_type: str = "application/json") -> dict:
    """A method to translate simple output example lists and dictionaries into FastAPI response dictionaries fit for
     use within the OpenAPI specifications and Swagger UI.

    Args:
        example_response (dict | list): An example response object (either a list or dict of items)
        content_type:                   A string holding the output content type.

    Returns:
        A FastAPI response dictionary fit for the OpenAPI specifications (and Swagger UI)

    """
    return {"content": {content_type: {"example": example_response}}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from typing import List

import pandas as pd

from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


def coords_to_pd_index(coords: List[GeoPosition]) -> pd.MultiIndex:
    return pd.MultiIndex.from_tuples([coord.get_WGS84() for coord in coords], names=("lat", "lon"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Utilities for handling KNMI datasets."""

import locale
import re
from datetime import datetime
from io import StringIO
from typing import List

import numpy as np
import pandas as pd
import requests
import xarray as xr
from bs4 import BeautifulSoup
from geopy.distance import great_circle
from loguru import logger

from weather_provider_api.routers.weather.sources.knmi.stations import (
    stations_actual,
    stations_actual_reversed,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


def find_closest_stn_list(
    stn_stations: pd.DataFrame, coords: List[GeoPosition]
) -> tuple[list[np.int64], list[np.int64], list[int]]:
    """A function that finds the closest stations to the locations in the given list of GeoPositions
    Args:
            stn_stations:   A Pandas Dataframe containing all the station data for the KNMI stations
            coords:         A list of GeoPositions with locations to find the nearest stations to.

    Returns:
            A list of coordinates for found stations
            A list of found stations in the same order
            A list of index numbers for the found in the same order
    """
    _stn_stations = stn_stations.copy(deep=True)

    # get list of relevant STNs, choose closest STN
    coords_stn = [_find_closest_stn_single(_stn_stations, coord) for coord in coords]
    stns = list(set(coords_stn))
    coords_stn_ind = [stns.index(x) for x in coords_stn]
    return coords_stn, stns, coords_stn_ind


def _find_closest_stn_single(stn_stations: pd.DataFrame, coord: GeoPosition) -> np.int64:
    """A function that finds the closest station to a single GeoPosition
    Args:
        stn_stations:   A Pandas Dataframe containing all the station data for the KNMI stations
        coord:          A GeoPosition to find the nearest station for
    Returns:
        A station number indicating its index in the supplied dataframe.
    """
    stn_stations["distance"] = stn_stations.apply(
        lambda x: great_circle((x["lat"], x["lon"]), coord.get_WGS84()).km, axis=1
    )

    # Find the stn with the lowest distance to the location
    min_ind = np.argmin(stn_stations["distance"].values)
    # Return the found stn
    return stn_stations.loc[min_ind, "STN"]


def download_actuele_waarnemingen_weather() -> xr.Dataset | None:
    """A function that downloads the current weather data from the KNMI site for the Actuele Waarnemingen model and returns it as a Xarray Dataset."""
    raw_ds = None

    try:
        knmi_site_response = requests.get("https://www.knmi.nl/nederland-nu/weer/waarnemingen", timeout=10)

        if knmi_site_response.ok:
            soup = BeautifulSoup(knmi_site_response.text, "lxml")
            table_wrp_div = soup.find("div", class_="table__wrp")
            if table_wrp_div is None:
                logger.error("Could not find div with class 'table__wrp' in KNMI Waarnemingen page.")
                raise IndexError("Could not find div with class 'table__wrp' in KNMI Waarnemingen page.")
            # Only parse the table inside the div
            tables = pd.read_html(StringIO(str(table_wrp_div)))
            if not tables:
                logger.error("No tables found inside 'table__wrp' div on KNMI Waarnemingen page.")
                raise IndexError("No tables found inside 'table__wrp' div on KNMI Waarnemingen page.")
            knmi_site_df = tables[0]

            column_translations = {
                "Station": "station",
                "Weer": "weather_description",
                "Temp (°C)": "temperature",
                "Chill (°C)": "wind_chill",
                "RV (%)": "relative_humidity",
                "Wind (bft)": "wind_direction",
                "Wind (m/s)": "wind_speed",
                "Wind (km/uur)": "wind_speed_kmu",
                "Zicht (m)": "visibility",
                "Druk (hPa)": "air_pressure",
            }

            # Rename to the conventional naming system used for the Weather Provider API
            for dictionary_item in knmi_site_df.columns.copy(deep=True):
                if dictionary_item in column_translations.keys():
                    knmi_site_df = knmi_site_df.rename(columns={dictionary_item: column_translations[dictionary_item]})
                else:
                    knmi_site_df = knmi_site_df.drop(dictionary_item, axis="columns")

            
            current_observation_moment = _retrieve_observation_moment(knmi_site_response.text)
            knmi_site_df["time"] = current_observation_moment
            knmi_site_df["time"] = knmi_site_df["time"].astype("datetime64[ns]")
            if "wind_direction" in knmi_site_df:
                knmi_site_df["wind_direction"] = knmi_site_df["wind_direction"].str.strip("\n")

            knmi_site_df["STN"] = knmi_site_df["station"].apply(lambda x: stations_actual_reversed[x.upper()])

            stations_actual_indexed = stations_actual.set_index("STN")

            knmi_site_df["lat"] = knmi_site_df["STN"].apply(lambda x: stations_actual_indexed.loc[x, "lat"])
            knmi_site_df["lon"] = knmi_site_df["STN"].apply(lambda x: stations_actual_indexed.loc[x, "lon"])
            knmi_site_df = knmi_site_df.drop("station", axis="columns")

            # Rebuild the index
            knmi_site_df = knmi_site_df.set_index(["time", "STN"])

            # The following fields get removed because they aren't currently used in this API
            if "weather_description" in knmi_site_df:
                knmi_site_df = knmi_site_df.drop("weather_description", axis="columns")
            if "wind_chill" in knmi_site_df:
                knmi_site_df = knmi_site_df.drop("wind_chill", axis="columns")

            # Generate a Xarray DataSet
            raw_ds = knmi_site_df.to_xarray()
        else:
            logger.error(
                f"Failed to retrieve KNMI Waarnemingen data. Received status code {knmi_site_response.status_code} when trying to access the site."
            )

    except (requests.exceptions.RequestException, IndexError) as expected_error:
        logger.exception(expected_error.__str__())
        raise expected_error
    except Exception as e:
        logger.exception(f"Unexpected error in download_actuele_waarnemingen_weather: {e}")
        raise e

    logger.info("Successfully retrieved KNMI Waarnemingen data.")
    return raw_ds


def _retrieve_observation_moment(html_body: str) -> datetime:
    """A function that extracts the observation date from the supplied html body of the Actuele Waarnemingen site.

    Args:
        html_body:  The text from the body downloaded from the Actuele Waarnemingen Site.

    Returns:
        Either the matching datetime if it can be extracted, or the current datetime if it cannot.
    """
    try:
        re_match = re.search(
            r"Waarnemingen\s(\d+\s\w+\s\d{4}\s\d{2}:\d{2})\suur",
            html_body,
        )
        if re_match:
            dt_str = re_match.group(1)
            current_locale = locale.getlocale(locale.LC_TIME)
            locale.setlocale(locale.LC_TIME, "dutch")
            dt = datetime.strptime(dt_str, "%d %B %Y %H:%M")
            locale.setlocale(locale.LC_TIME, current_locale)
            return dt
    except locale.Error:
        logger.warning(
            "No locale could be determined for KNMI Waarnemingen datetime transformation. "
            "Using local datetime instead."
        )
        return datetime.now()
    except Exception as e:
        logger.error(
            "An unknown exception occurred while retrieving the KNMI Waarnemingen datetime for use in the response."
            f" Using local datetime instead. Error: {e}"
        )
        return datetime.now()

    current_locale = locale.getlocale(locale.LC_TIME)
    locale.setlocale(locale.LC_TIME, "dutch")
    dt = datetime.now()
    locale.setlocale(locale.LC_TIME, current_locale)
    return dt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0


from fastapi import FastAPI
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request

from weather_provider_api.app_version import APP_VERSION
from weather_provider_api.config import APP_CONFIG


def initialize_header_metadata(application: FastAPI):
    """Method that attaches the customized Metadata Header method that adds extra metadata.

    Args:
        application: The FastAPI application to attach the custom method to.

    Returns:
        Nothing. The FastAPI application itself is updated.
    """

    async def add_metadata_headers(request: Request, call_next):
        response = await call_next(request)
        response.headers["X-App-Version"] = APP_VERSION
        response.headers["X-App-Valid-Till"] = APP_CONFIG["base"]["expiration_date"]

        if APP_CONFIG["maintainer"]["show_info"]:
            response.headers["X-Maintainer"] = APP_CONFIG["maintainer"]["name"]
            response.headers["X-Maintainer-Email"] = APP_CONFIG["maintainer"]["email_address"]

        return response

    application.add_middleware(BaseHTTPMiddleware, dispatch=add_metadata_headers)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from fastapi import FastAPI

from weather_provider_api.config import APP_CONFIG, APP_SERVER
from weather_provider_api.routers.weather.api_view_v1 import app as weather_router

app = FastAPI(
    version=APP_CONFIG["api_v1"]["implementation"],
    title="Weather API (v1)",
    root_path="/api/v1",
    servers=[{"url": f"{APP_SERVER}/api/v1"}],
    description="The v1 endpoint interface for the Weather Provider API",
    contact={
        "name": APP_CONFIG["maintainer"]["name"],
        "email": APP_CONFIG["maintainer"]["email_address"],
    },
)

app.openapi_version = "3.0.2"

app.include_router(weather_router, prefix="/api/v1/weather")
app.openapi()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from abc import ABCMeta, abstractmethod
from typing import List, Optional

import numpy as np
import xarray as xr

from weather_provider_api.core.initializers.exception_handling import (
    NOT_IMPLEMENTED_ERROR,
)
from weather_provider_api.routers.weather.api_models import OutputUnit
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class WeatherModelBase(metaclass=ABCMeta):
    """Base class for all Weather Models. All new models should use this base class!"""

    def __init__(self):
        self.to_si = None
        self.to_human = None
        self.human_to_model_specific = None

    @abstractmethod
    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: Optional[np.datetime64],
        end: Optional[np.datetime64],
        weather_factors: List[str] = None,
    ) -> xr.Dataset:  # pragma: no cover
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    @abstractmethod
    def is_async(self):  # pragma: no cover
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    def convert_names_and_units(self, weather_data: xr.Dataset, unit: OutputUnit):
        """Function to convert all names and units in a dataset according to the required translations set in the
            to_xxxx values for the model itself
        Args:
            weather_data:   A Xarray Dataset containing
            unit:           The requested output unit format

        Returns:
            The same dataset, but with values altered to match the requested output unit format
        """
        if unit == OutputUnit.original:
            return weather_data

        if unit == OutputUnit.si:
            conversion_dict = self.to_si
        elif unit == OutputUnit.human:
            conversion_dict = self.to_human
        else:
            raise TypeError("Invalid OutputUnit")

        data_vars = list(weather_data.data_vars)
        for var_name in data_vars:
            if var_name not in data_vars or var_name not in conversion_dict:
                continue

            new_name = var_name
            if "name" in conversion_dict[var_name]:
                new_name = conversion_dict[var_name]["name"]

            new_data = weather_data[var_name]
            if "convert" in conversion_dict[var_name]:
                dtype_value = new_data.dtype
                converter: callable = conversion_dict[var_name]["convert"]
                new_data = converter(weather_data[var_name]).astype(dtype_value)

            weather_data = weather_data.drop_vars(var_name)
            weather_data[new_name] = new_data

        return weather_data

    @staticmethod
    def _create_reverse_lookup(conversion_dict):  # pragma: no cover
        reverse = dict()

        for k, v in conversion_dict.items():
            if "name" in v:
                reverse[v["name"]] = k

        return reverse

    @abstractmethod
    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    @staticmethod
    def celsius_to_kelvin(x):  # pragma: no cover
        return x + 273.15

    @staticmethod
    def kelvin_to_celsius(x):  # pragma: no cover
        return x - 273.15

    def tenth_celsius_to_kelvin(self, x):  # pragma: no cover
        return self.celsius_to_kelvin(self.normalize_tenths(x))

    @staticmethod
    def normalize_tenths(x):  # pragma: no cover
        return x / 10

    @staticmethod
    def no_conversion(x):  # pragma: no cover
        return x

    @staticmethod
    def percentage_to_frac(x):  # pragma: no cover
        return x / 100

    @staticmethod
    def kmh_to_ms(x):  # pragma: no cover
        return x / 3.6

    @staticmethod
    def dutch_wind_direction_to_degrees(xs):
        wind_directions = [
            "NNO",
            "NO",
            "ONO",
            "O",
            "OZO",
            "ZO",
            "ZZO",
            "Z",
            "ZZW",
            "ZW",
            "WZW",
            "W",
            "WNW",
            "NW",
            "NNW",
            "N",
        ]

        def dutch_wind_direction_to_degrees_single(x):
            if x in wind_directions:
                return (wind_directions.index(x) + 1) * 22.5
            else:
                return None

        return np.frompyfunc(dutch_wind_direction_to_degrees_single, 1, 1)(xs)

    @staticmethod
    def knmi_visibility_class_to_meter_estimate(xs):
        """Function to transform KNMI visibility class values to an estimate of meters visibility
        Args:
            xs:     The visibility class value to be interpreted

        Returns:
            A numeric value containing an estimate of the meters of visibility matching the given visibility class value

        """

        def knmi_visibility_class_to_meter_estimate_single(x):
            if x < 50:
                return x * 100 + 50
            elif x == 50:
                return 5500
            elif x < 80:
                return (x - 50) * 1000 + 500
            elif x < 89:
                return (x - 80) * 5000 + 30000 + 2500
            else:
                # Even though the value 0f 88 results in 72500 as a return value, the final return of 70000 is correct.
                # From 80 up to 88 each return represents a value range of 5000, and for 88 this is 70000 - 75000,
                # represented by 72500. From 89 up the value is given as "more than 70000", which is why 70000 is
                # returned from then on.
                return 70000

        return np.frompyfunc(knmi_visibility_class_to_meter_estimate_single, 1, 1)(xs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""KNMI day part prediction data fetcher."""

import copy
from datetime import datetime
from typing import List, Optional

import numpy as np
import requests
import xarray as xr
from dateutil.relativedelta import relativedelta
from loguru import logger
from pytz import UTC

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.stations import (
    stations_prediction,
)
from weather_provider_api.routers.weather.sources.knmi.utils import (
    find_closest_stn_list,
)
from weather_provider_api.routers.weather.utils.date_helpers import (
    validate_begin_and_end,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.pandas_helpers import coords_to_pd_index


class PluimModel(WeatherModelBase):
    """A Weather Model that incorporates the:
        KNMI Pluim
    dataset into the Weather Provider API
    """

    def __init__(self):
        super().__init__()
        self.id = "pluim"
        logger.debug(f"Initializing weather model [{self.id}]")

        self.name = "ECMWF pluim"
        self.version = ""
        self.url = "https://www.knmi.nl/nederland-nu/weer/waarschuwingen-en-verwachtingen/weer-en-klimaatpluim"
        self.predictive = True
        self.time_step_size_minutes = 720
        self.num_time_steps = 30
        self.description = (
            "Predictions for the coming 15 days, current included, with two predictions made for each day."
        )
        self.async_model = False

        self.to_si = {
            "wind_speed": {
                "name": "wind_speed",
                "convert": self.kmh_to_ms,  # km/h -> m/s
                "code": 11012,
            },
            "wind_direction": {
                "name": "wind_direction",
                "convert": self.no_conversion,  # 360N, 270W, ...
                "code": 11011,
            },
            "short_time_wind_speed": {
                "name": "wind_speed_max",
                "convert": self.kmh_to_ms,  # km/h -> m/s
                "code": 11041,
            },
            "temperature": {
                "name": "temperature",
                "convert": self.celsius_to_kelvin,  # degree C -> Kelvin
                "code": 99999,
            },
            "precipitation": {
                "name": "precipitation",
                "convert": lambda x: x / 1000,  # mm -> m
                "code": 13021,
            },
            "precipitation_sum": {
                "name": "precipitation_sum",
                "convert": lambda x: x / 1000,  # mm -> m
                "code": 13011,
            },
            "cape": {
                "name": "cape",
                "convert": lambda x: x / 1000,
                "code": 13241,
            },  # J/kg
        }

        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["temperature"]["convert"] = self.no_conversion  # C -> C
        self.to_human["precipitation"]["convert"] = self.no_conversion  # mm
        self.to_human["precipitation_sum"]["convert"] = self.no_conversion  # mm

        self.human_to_model_specific = self._create_reverse_lookup(self.to_si)

        logger.debug(f"Weather model [{self.id}] initialized successfully")

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: datetime,
        end: datetime,
        weather_factors: List[str] = None,
        **_kwargs,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested Daggegevens weather data from the KNMI site
            and returns it as a Xarray Dataset.
            (Though this model downloads from a specific download url, the question remains whether this source is also
            listed on the new KNMI Data Platform)

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.
        """
        # Test and account for invalid datetime timeframes or input
        begin, end = validate_begin_and_end(begin, end, datetime.utcnow(), datetime.utcnow() + relativedelta(days=15))

        # get list of relevant STNs, choose closest STN
        _, stns, coords_stn_ind = find_closest_stn_list(stations_prediction, coords)

        # load default weather factors if unspecified
        if weather_factors is None:
            weather_factors = self.to_si.keys()

        # download the weather factors for all stations
        ds = self._download_weather(coords, coords_stn_ind, stns, weather_factors)

        ds = self._select_weather_from_given_period(ds, begin, end)
        ds = ds.dropna("time", how="all")  # Dropping any times that only carry NaN values

        return ds

    @staticmethod
    def _select_weather_from_given_period(ds: xr.Dataset, begin: datetime, end: datetime):
        """A function that filters the given Xarray Dataset to only the requested period.

        Args:
            ds:     An Xarray Dataset to be filtered
            begin:  A datetime containing the start of the period to filter.
            end:    A datetime containing the end of the period to filter.

        Returns:
            An Xarray Dataset containing all the data from the original dataset that matches the given period.
        """
        begin, end = validate_begin_and_end(
            begin,
            end,
            datetime.today().replace(hour=0, minute=0, second=0),
            datetime.today().replace(hour=0, minute=0, second=0) + relativedelta(days=15),
        )
        ds = ds.sel(time=slice(begin.astimezone(UTC).replace(tzinfo=None), end.astimezone(UTC).replace(tzinfo=None)))
        return ds

    def is_async(self):  # pragma: no cover
        return self.async_model

    def _download_weather(
        self,
        coordinates: List[GeoPosition],
        coords_stn_ind,
        stns,
        weather_factors: List[str],
    ):
        """A function that downloads the requested weather data, factor by factor.

        Args:
            coordinates:        A list of GeoPositions containing the coordinates to request data for.
            coords_stn_ind:     A list of station coordinates in order
            stns:               A list of stations by ID in the same order
            weather_factors:    A list of weather factors to request the data for (string based)

        Returns:
            An Xarray Dataset containing the requested factors for the requested period and location(s).
        """
        arr_dict = {}
        ds = xr.Dataset()
        for weather_factor in weather_factors:
            try:
                # Convert the factor to its request code
                code = self.to_si[weather_factor]["code"]
            except KeyError:
                # Requested factor is unknown: skip
                continue

            # Download the requested factor
            timeline, value = self._download_single_factor(stns, code)

            # add values
            arr_dict[weather_factor] = xr.DataArray(
                data=value[:, coords_stn_ind],
                dims=["time", "coord"],
                coords={"time": timeline, "coord": coords_to_pd_index(coordinates)},
                name=weather_factor,
            )
            ds = xr.merge(arr_dict.values(), join="outer")
        ds = ds.unstack("coord")
        return ds

    @staticmethod
    def _download_single_stn_factor(stn: int, factor: int):
        """A function to download a single factor for a single station
        Args:
            stn:    The station (ID) to download the factor for
            factor: The weather factor to download
        Returns:
            A list of datetime64 items holding the times for the factors, and a matching list of values holding
            values belonging to the requested factor during those times.
        """
        base_url = (
            f"https://cdn.knmi.nl/knmi/json/page/weer/waarschuwingen_verwachtingen/ensemble/iPluim/{stn}_{factor}.json"
        )

        response = requests.get(base_url)
        if response.status_code != 200:
            raise requests.HTTPError(
                "Failed to retrieve data from the KNMI website",
                response.status_code,
                base_url,
            )

        series = response.json()["series"]
        # only retrieve prediction
        prediction_data = [x for x in series if "Verwacht" in x["name"]][0]["data"]
        if isinstance(prediction_data[0], dict):
            # get timeline
            timeline = np.array([x["x"] for x in prediction_data]).astype("datetime64[ms]")
            # get value
            value = np.array([x["y"] for x in prediction_data], dtype=np.float64)
        else:
            # get timeline
            timeline = np.array([x[0] for x in prediction_data]).astype("datetime64[ms]")
            # get value
            value = np.array([x[1] for x in prediction_data], dtype=np.float64)

        return timeline, value

    def _download_single_factor(self, stns: list, factor: int):
        """A function that downloads a single factor for all the given stations
        Args:
            stns:   A list of stations to download the factor for.
            factor: The weather factor to download.

        Returns:
            A list holding a timeline in datetime64 items as returned from the download requests, matching the request,
            and a list holding the values for the requested factor for all the stations in the order:
                Station, Timeline
        """
        values = []
        timeline = []
        for stn in stns:
            timeline, value = self._download_single_stn_factor(stn, factor)
            values.append(value)
        return timeline, np.stack(values, axis=1)

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()

            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Exception Handling

This module holds the API's exception handler.
"""

from fastapi import FastAPI
from loguru import logger
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse

from weather_provider_api.config import APP_CONFIG


async def handle_http_exception(request: Request, exc: StarletteHTTPException) -> JSONResponse:
    headers = getattr(exc, "headers", None)

    body = {"detail": exc.detail, "request": str(request.url)}
    if APP_CONFIG["maintainer"]["show_info"]:
        body["maintainer"] = APP_CONFIG["maintainer"]["name"]
        body["maintainer_email"] = APP_CONFIG["maintainer"]["email_address"]

    return JSONResponse(body, status_code=exc.status_code, headers=headers)


def initialize_exception_handler(application: FastAPI):
    """The method that attaches the customized exception handling method to a FastAPI application.

    Args:
        application:    The FastAPI application to attach the custom exception handler to.

    Returns:
        Nothing. The FastAPI application itself is updated.

    Notes:
        This method assumes that the FastAPI application given has a [title] set.

    Todo:
        Evaluate the dependency on [title] and [root_path] parameters being set for FastAPI applications in ALL
         initializers. By either extending the base class to enforce these or improving the code to not be dependent on
         these parameters, we can eradiate code smell and chances at Exceptions caused by not having these parameters.
    """
    application.add_exception_handler(StarletteHTTPException, handler=handle_http_exception)
    logger.info(f"Attached the Exception Handler to the application ({application.title})...")


NOT_IMPLEMENTED_ERROR = "This method is abstract and should be overridden."



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0


from weather_provider_api.routers.weather.sources.cds.client.era5land_repository import (
    ERA5LandRepository,
)


def main():
    era5land_repo = ERA5LandRepository()
    era5land_repo.purge_repository()


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from typing import Any

from fastapi import HTTPException
from pydantic import BaseModel


class ExceptionResponseModel(BaseModel):
    """Class only used to relay the output for the HTTP Exception classes to the OpenAPI specification and
    Swagger UI.
    """

    detail: str


class APIExpiredException(HTTPException):
    def __init__(self, detail: Any = None):
        self.detail = (
            detail
            or "This API has passed it's expiry date and should be revalidated. Please contact the API maintainer."
        )
        self.status_code = 404



import json

from weather_provider_api.core.application import WPLA_APPLICATION
from weather_provider_api.versions.v2 import app as v2_app



def generate_openapi_spec():
    """Generate OpenAPI specification and save it to a file."""
    spec = v2_app.openapi()
    with open("openapi.json", "w") as f:
        f.write(json.dumps(spec))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.routers.weather.base_models.source import WeatherSourceBase
from weather_provider_api.routers.weather.sources.cds.models.era5sl import ERA5SLModel


class CDS(WeatherSourceBase):
    def __init__(self):
        model_instances = [
            ERA5SLModel(),
        ]

        self.id = "cds"
        self.name = "Climate Data Store"
        self.url = "https://cds.climate.copernicus.eu/"
        self._models = None
        self._async_models = None

        self.setup_models(model_instances)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""The list of stations used in various KNMI datasets."""

from collections import namedtuple

import numpy as np
import pandas as pd

_Station = namedtuple("_Station", ["number", "longitude", "latitude", "altitude", "name"])


_stations = {
    8: _Station(number=8, longitude=5.528, latitude=52.453, altitude=-4.2, name="LELYSTAD"),
    13: _Station(
        number=13,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="WISSEKERKE ???",
    ),
    16: _Station(
        number=16,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="HEINKENSZAND ???",
    ),
    17: _Station(number=17, longitude=3.892, latitude=51.529, altitude=0.7, name="WILHELMINADORP"),
    18: _Station(
        number=18,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="PHILIPPINE ???",
    ),
    19: _Station(
        number=19,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="NIEUWE TONGE ???",
    ),
    20: _Station(number=20, longitude=6.7, latitude=51.969, altitude=34.1, name="WINTERSWIJK"),
    28: _Station(number=28, longitude=5.82, latitude=52.961, altitude=-0.4, name="JOURE"),
    29: _Station(number=29, longitude=5.044, latitude=52.65, altitude=999.9, name="HOORN (NH)"),
    32: _Station(number=32, longitude=4.526, latitude=51.576, altitude=1.2, name="OUDENBOSCH"),
    33: _Station(number=33, longitude=5.677, latitude=51.549, altitude=13.2, name="GEMERT"),
    34: _Station(number=34, longitude=6.887, latitude=52.785, altitude=24.1, name="EMMEN"),
    35: _Station(
        number=35,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="HAAMSTEDE ???",
    ),
    36: _Station(
        number=36,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="NUMANSDORP ???",
    ),
    37: _Station(
        number=37,
        longitude=5.675,
        latitude=51.976,
        altitude=2.0,
        name="WAGENINGEN (HAARWEG)",
    ),
    38: _Station(
        number=38,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="NAALDWIJK ???",
    ),
    39: _Station(
        number=39,
        longitude=4.156,
        latitude=51.988,
        altitude=999.9,
        name="H. VAN HOLLAND(MOLENPAD)",
    ),
    40: _Station(
        number=40,
        longitude=5.837,
        latitude=51.856,
        altitude=999.9,
        name="NIJMEGEN (RADBOUD)",
    ),
    41: _Station(number=41, longitude=5.526, latitude=52.458, altitude=-4.0, name="LELYSTAD"),
    108: _Station(
        number=108,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="HELLEVOETSLUIS ???",
    ),
    122: _Station(number=122, longitude=6.303, latitude=52.158, altitude=9.0, name="ALMEN"),
    125: _Station(number=125, longitude=5.054, latitude=51.784, altitude=1.4, name="ANDEL"),
    126: _Station(number=126, longitude=4.079, latitude=51.596, altitude=1.0, name="ST ANNALAND"),
    127: _Station(number=127, longitude=3.314, latitude=47.975, altitude=999.9, name="BUCHTEN ???"),
    128: _Station(number=128, longitude=6.443, latitude=52.593, altitude=6.0, name="DEDEMSVAART"),
    129: _Station(number=129, longitude=5.78, latitude=52.712, altitude=-3.7, name="EMMELOORD"),
    130: _Station(number=130, longitude=3.314, latitude=47.975, altitude=999.9, name="EPEN ???"),
    133: _Station(
        number=133,
        longitude=5.342,
        latitude=53.398,
        altitude=-1.0,
        name="HOORN (TERSCHELLING)",
    ),
    135: _Station(number=135, longitude=5.338, latitude=53.07, altitude=3.0, name="KORNWERDERZAND"),
    138: _Station(number=138, longitude=5.676, latitude=50.843, altitude=999.9, name="MAASTRICHT"),
    139: _Station(number=139, longitude=5.121, latitude=52.458, altitude=-0.8, name="MARKEN"),
    142: _Station(
        number=142,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="OUDE WETERING ???",
    ),
    143: _Station(number=143, longitude=4.393, latitude=51.855, altitude=0.2, name="POORTUGAAL"),
    147: _Station(number=147, longitude=5.611, latitude=52.658, altitude=999.9, name="URK"),
    148: _Station(number=148, longitude=6.191, latitude=51.356, altitude=30.0, name="VENLO"),
    152: _Station(number=152, longitude=3.558, latitude=51.356, altitude=1.2, name="SCHOONDIJKE"),
    153: _Station(number=153, longitude=5.356, latitude=52.886, altitude=2.6, name="STAVOREN"),
    158: _Station(
        number=158,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="SPIJK (GR.) ???",
    ),
    159: _Station(number=159, longitude=4.057, latitude=51.251, altitude=2.4, name="KAPELLEBRUG"),
    161: _Station(number=161, longitude=5.975, latitude=53.391, altitude=999.9, name="TERNAARD"),
    162: _Station(
        number=162,
        longitude=5.222,
        latitude=52.418,
        altitude=999.9,
        name="OOSTVAARDERSDIEP",
    ),
    163: _Station(
        number=163,
        longitude=5.476,
        latitude=52.263,
        altitude=0.2,
        name="NIJKERK (SLUIS)",
    ),
    164: _Station(number=164, longitude=5.909, latitude=51.092, altitude=34.1, name="ECHT"),
    165: _Station(
        number=165,
        longitude=3.314,
        latitude=47.975,
        altitude=999.9,
        name="KAAP HOOFD ???",
    ),
    166: _Station(
        number=166,
        longitude=6.238,
        latitude=53.583,
        altitude=4.0,
        name="SCHIERMONNIKOOG",
    ),
    167: _Station(number=167, longitude=6.85, latitude=53.381, altitude=999.9, name="BIERUM"),
    168: _Station(number=168, longitude=5.906, latitude=50.764, altitude=155.0, name="EPEN"),
    169: _Station(number=169, longitude=5.304, latitude=53.392, altitude=0.4, name="FORMERUM"),
    170: _Station(
        number=170,
        longitude=5.715,
        latitude=50.794,
        altitude=50.0,
        name="OOST MAARLAND",
    ),
    200: _Station(number=200, longitude=4.355, latitude=52.041, altitude=999.9, name="YPENBURG"),
    201: _Station(number=201, longitude=2.936, latitude=54.327, altitude=42.7, name="D15-FA-1"),
    202: _Station(number=202, longitude=4.362, latitude=52.206, altitude=9999.9, name="MP KATWIJK"),
    203: _Station(number=203, longitude=3.342, latitude=52.361, altitude=41.8, name="P11-B"),
    204: _Station(number=204, longitude=3.628, latitude=53.271, altitude=41.8, name="K14-FA-1C"),
    205: _Station(number=205, longitude=3.81, latitude=55.401, altitude=48.4, name="A12-CPP"),
    206: _Station(number=206, longitude=4.012, latitude=54.118, altitude=43.4, name="F16-A"),
    207: _Station(number=207, longitude=4.961, latitude=53.616, altitude=44.0, name="L9-FF-1"),
    208: _Station(number=208, longitude=5.942, latitude=53.493, altitude=40.5, name="AWG-1"),
    209: _Station(number=209, longitude=4.518, latitude=52.465, altitude=0.0, name="IJMOND"),
    210: _Station(number=210, longitude=4.43, latitude=52.171, altitude=-0.2, name="VALKENBURG"),
    211: _Station(number=211, longitude=2.945, latitude=53.825, altitude=45.7, name="J6-A"),
    212: _Station(number=212, longitude=4.151, latitude=52.919, altitude=50.9, name="HOORN-A"),
    214: _Station(number=214, longitude=6.042, latitude=54.038, altitude=42.5, name="Gemini"),
    215: _Station(number=215, longitude=4.437, latitude=52.141, altitude=-1.1, name="VOORSCHOTEN"),
    225: _Station(number=225, longitude=4.555, latitude=52.463, altitude=4.4, name="IJMUIDEN"),
    227: _Station(number=227, longitude=4.655, latitude=52.774, altitude=0.5, name="PETTEN"),
    229: _Station(number=229, longitude=4.72, latitude=52.996, altitude=10.0, name="TEXELHORS MM"),
    230: _Station(number=230, longitude=4.759, latitude=52.967, altitude=999.9, name="DEN HELDER"),
    235: _Station(number=235, longitude=4.781, latitude=52.928, altitude=1.2, name="DE KOOY"),
    239: _Station(number=239, longitude=4.696, latitude=54.855, altitude=50.6, name="F3-FB-1"),
    240: _Station(number=240, longitude=4.79, latitude=52.318, altitude=-3.3, name="SCHIPHOL"),
    242: _Station(number=242, longitude=4.921, latitude=53.241, altitude=10.8, name="VLIELAND"),
    244: _Station(number=244, longitude=5.044, latitude=52.65, altitude=999.9, name="HOORN (NH)"),
    247: _Station(number=247, longitude=4.555, latitude=52.422, altitude=18.0, name="BLOEMENDAAL"),
    248: _Station(number=248, longitude=5.174, latitude=52.634, altitude=0.8, name="WIJDENES"),
    249: _Station(number=249, longitude=4.979, latitude=52.644, altitude=-2.4, name="BERKHOUT"),
    250: _Station(number=250, longitude=5.147, latitude=53.36, altitude=31.4, name="TERSCHELLING"),
    251: _Station(
        number=251,
        longitude=5.346,
        latitude=53.392,
        altitude=0.7,
        name="HOORN (TERSCHELLING)",
    ),
    252: _Station(number=252, longitude=3.22, latitude=53.219, altitude=37.7, name="K13"),
    253: _Station(number=253, longitude=2.067, latitude=56.4, altitude=999.9, name="AUK-ALFA"),
    254: _Station(
        number=254,
        longitude=4.296,
        latitude=52.274,
        altitude=999.9,
        name="MEETPOST NOORDWIJK",
    ),
    255: _Station(
        number=255,
        longitude=1.16,
        latitude=61.234,
        altitude=999.9,
        name="NORTH CORMORANT",
    ),
    257: _Station(number=257, longitude=4.603, latitude=52.506, altitude=8.5, name="WIJK AAN ZEE"),
    258: _Station(number=258, longitude=5.401, latitude=52.649, altitude=7.3, name="HOUTRIBDIJK"),
    260: _Station(number=260, longitude=5.18, latitude=52.1, altitude=1.9, name="DE BILT"),
    263: _Station(
        number=263,
        longitude=5.177,
        latitude=52.101,
        altitude=2.0,
        name="DE BILT (TEST)",
    ),
    265: _Station(number=265, longitude=5.274, latitude=52.13, altitude=13.9, name="SOESTERBERG"),
    266: _Station(number=266, longitude=5.338, latitude=53.07, altitude=3.0, name="KORNWERDERZAND"),
    267: _Station(number=267, longitude=5.384, latitude=52.898, altitude=-1.3, name="STAVOREN"),
    268: _Station(number=268, longitude=5.436, latitude=52.531, altitude=999.9, name="HOUTRIB"),
    269: _Station(number=269, longitude=5.52, latitude=52.458, altitude=-3.7, name="LELYSTAD"),
    270: _Station(number=270, longitude=5.752, latitude=53.224, altitude=1.2, name="LEEUWARDEN"),
    271: _Station(number=271, longitude=5.356, latitude=52.888, altitude=2.6, name="STAVOREN"),
    272: _Station(number=272, longitude=5.85, latitude=52.616, altitude=3.6, name="RAMSPOL"),
    273: _Station(number=273, longitude=5.888, latitude=52.703, altitude=-3.3, name="MARKNESSE"),
    275: _Station(number=275, longitude=5.873, latitude=52.056, altitude=48.2, name="DEELEN"),
    277: _Station(number=277, longitude=6.2, latitude=53.413, altitude=2.9, name="LAUWERSOOG"),
    278: _Station(number=278, longitude=6.259, latitude=52.435, altitude=3.6, name="HEINO"),
    279: _Station(number=279, longitude=6.574, latitude=52.75, altitude=15.8, name="HOOGEVEEN"),
    280: _Station(number=280, longitude=6.585, latitude=53.125, altitude=5.2, name="EELDE"),
    283: _Station(number=283, longitude=6.657, latitude=52.069, altitude=29.1, name="HUPSEL"),
    285: _Station(number=285, longitude=6.399, latitude=53.575, altitude=0.0, name="HUIBERTGAT"),
    286: _Station(number=286, longitude=7.15, latitude=53.196, altitude=-0.2, name="NIEUW BEERTA"),
    290: _Station(number=290, longitude=6.891, latitude=52.274, altitude=34.8, name="TWENTHE"),
    300: _Station(
        number=300,
        longitude=2.567,
        latitude=51.65,
        altitude=999.9,
        name="LICHTSCHIP NOORDHINDER",
    ),
    308: _Station(number=308, longitude=3.379, latitude=51.381, altitude=0.0, name="CADZAND"),
    310: _Station(number=310, longitude=3.596, latitude=51.442, altitude=8.0, name="VLISSINGEN"),
    311: _Station(number=311, longitude=3.672, latitude=51.379, altitude=0.0, name="HOOFDPLAAT"),
    312: _Station(number=312, longitude=3.622, latitude=51.768, altitude=0.0, name="OOSTERSCHELDE"),
    313: _Station(
        number=313,
        longitude=3.242,
        latitude=51.505,
        altitude=0.0,
        name="VLAKTE V.D. RAAN",
    ),
    315: _Station(number=315, longitude=3.998, latitude=51.447, altitude=0.0, name="HANSWEERT"),
    316: _Station(number=316, longitude=3.694, latitude=51.657, altitude=0.0, name="SCHAAR"),
    319: _Station(number=319, longitude=3.861, latitude=51.226, altitude=1.7, name="WESTDORPE"),
    320: _Station(
        number=320,
        longitude=3.67,
        latitude=51.927,
        altitude=22.0,
        name="LICHTEILAND GOEREE",
    ),
    321: _Station(number=321, longitude=3.275, latitude=51.999, altitude=19.0, name="EUROPLATFORM"),
    323: _Station(
        number=323,
        longitude=3.884,
        latitude=51.527,
        altitude=1.4,
        name="WILHELMINADORP",
    ),
    324: _Station(number=324, longitude=4.006, latitude=51.596, altitude=0.0, name="STAVENISSE"),
    325: _Station(number=325, longitude=3.931, latitude=51.654, altitude=-4.0, name="ZIERIKZEE"),
    328: _Station(number=328, longitude=3.821, latitude=51.668, altitude=999.9, name="ROGGENPLAAT"),
    330: _Station(
        number=330,
        longitude=4.122,
        latitude=51.992,
        altitude=11.9,
        name="HOEK VAN HOLLAND",
    ),
    331: _Station(number=331, longitude=4.193, latitude=51.48, altitude=0.0, name="THOLEN"),
    340: _Station(number=340, longitude=4.342, latitude=51.449, altitude=19.2, name="WOENSDRECHT"),
    343: _Station(
        number=343,
        longitude=4.313,
        latitude=51.893,
        altitude=3.5,
        name="R'DAM-GEULHAVEN",
    ),
    344: _Station(number=344, longitude=4.447, latitude=51.962, altitude=-4.3, name="ROTTERDAM"),
    348: _Station(number=348, longitude=4.926, latitude=51.97, altitude=-0.7, name="CABAUW"),
    350: _Station(number=350, longitude=4.936, latitude=51.566, altitude=14.9, name="GILZE-RIJEN"),
    356: _Station(number=356, longitude=5.146, latitude=51.859, altitude=0.7, name="HERWIJNEN"),
    370: _Station(number=370, longitude=5.377, latitude=51.451, altitude=22.6, name="EINDHOVEN"),
    375: _Station(number=375, longitude=5.707, latitude=51.659, altitude=22.0, name="VOLKEL"),
    377: _Station(number=377, longitude=5.763, latitude=51.198, altitude=30.0, name="ELL"),
    379: _Station(number=379, longitude=5.7, latitude=50.794, altitude=42.0, name="OOST-MAARLAND"),
    380: _Station(number=380, longitude=5.762, latitude=50.906, altitude=114.3, name="MAASTRICHT"),
    385: _Station(number=385, longitude=5.933, latitude=51.55, altitude=9999.9, name="DE PEEL"),
    391: _Station(number=391, longitude=6.197, latitude=51.498, altitude=19.5, name="ARCEN"),
    392: _Station(number=392, longitude=6.056, latitude=51.487, altitude=21.88, name="HORST"),
    550: _Station(number=550, longitude=3.22, latitude=53.218, altitude=999.9, name="K13"),
    551: _Station(number=551, longitude=2.066, latitude=56.4, altitude=999.9, name="AUK"),
    552: _Station(number=552, longitude=3.213, latitude=56.547, altitude=999.9, name="EKOFISK"),
    553: _Station(
        number=553,
        longitude=3.276,
        latitude=51.999,
        altitude=999.9,
        name="EUROPLATFORM",
    ),
    554: _Station(
        number=554,
        longitude=3.592,
        latitude=51.681,
        altitude=999.9,
        name="MEETPOST NOORDWIJK",
    ),
    603: _Station(number=603, longitude=6.189, latitude=52.311, altitude=999.9, name="DIEPENVEEN"),
    604: _Station(number=604, longitude=6.262, latitude=52.435, altitude=1.4, name="HERWIJNEN"),
    605: _Station(number=605, longitude=6.196, latitude=53.409, altitude=3.0, name="LAUWERSOOG"),
    609: _Station(
        number=609,
        longitude=4.318,
        latitude=51.891,
        altitude=999.9,
        name="R'DAM-GEULHAVEN",
    ),
    614: _Station(number=614, longitude=6.92, latitude=53.337, altitude=999.9, name="EEMSHAVEN"),
    615: _Station(number=615, longitude=4.296, latitude=52.274, altitude=12.0, name="HOOGEVEEN"),
    616: _Station(
        number=616,
        longitude=4.907,
        latitude=52.367,
        altitude=999.9,
        name="AMSTERDAM (COENHAVEN)",
    ),
    617: _Station(number=617, longitude=6.196, latitude=51.498, altitude=19.0, name="ARCEN"),
    995: _Station(
        number=995,
        longitude=3.596,
        latitude=51.442,
        altitude=8.0,
        name="VLISSINGEN (TEST)",
    ),
    998: _Station(
        number=998,
        longitude=5.177,
        latitude=52.101,
        altitude=2.0,
        name="DE BILT (TEST)",
    ),
    999: _Station(
        number=999,
        longitude=6.575,
        latitude=52.75,
        altitude=15.6,
        name="HOOGEVEEN TEST",
    ),
    917536001: _Station(
        number=917536001,
        longitude=5.762,
        latitude=50.905,
        altitude=113,
        name="MAASTRICHT-AACHEN AIRPORT",
    ),
}


# convert to data frame
_stations_df = pd.DataFrame(_stations).transpose()
_stations_df.index.name = "STN"
_stations_df.columns = ["STN", "lon", "lat", "alt", "name"]


# filter on prediction and models
def filter_stations(stns):
    filtered_stations: pd.DataFrame = _stations_df.copy(deep=True)
    filtered_stations.set_index("STN", inplace=True)
    filtered_stations = filtered_stations.loc[stns, :]
    filtered_stations.reset_index(level=0, inplace=True)
    return filtered_stations


# filter
_stations_prediction_idx = np.array([260, 235, 280, 270, 380, 240, 290, 310], dtype=np.int64)

_stations_history_idx = np.array(
    [
        209,
        215,
        225,
        235,
        240,
        242,
        248,
        249,
        251,
        257,
        258,
        260,
        267,
        269,
        270,
        273,
        275,
        277,
        278,
        279,
        280,
        283,
        285,
        286,
        290,
        308,
        310,
        312,
        313,
        315,
        316,
        319,
        323,
        324,
        330,
        331,
        340,
        343,
        344,
        348,
        350,
        356,
        370,
        375,
        377,
        380,
        391,
        392,
    ],
    dtype=np.int64,
)

_stations_actual_idx = np.array(
    [
        8,
        17,
        153,
        215,
        225,
        229,
        230,
        240,
        242,
        249,
        250,
        257,
        258,
        260,
        270,
        273,
        275,
        277,
        278,
        280,
        283,
        286,
        290,
        310,
        319,
        330,
        340,
        344,
        348,
        350,
        356,
        370,
        375,
        377,
        391,
        392,
        615,
        917536001,
    ],
    dtype=np.int64,
)


stations_prediction = filter_stations(_stations_prediction_idx)
stations_history = filter_stations(_stations_history_idx)
stations_actual = filter_stations(_stations_actual_idx)

stations_actual_reversed = {s["name"]: s["STN"] for s in stations_actual.reset_index().to_dict(orient="records")}
stations_actual_reversed["TWENTE"] = stations_actual_reversed["TWENTHE"]
stations_actual_reversed["TEXELHORS"] = stations_actual_reversed["TEXELHORS MM"]
stations_actual_reversed["GILZE RIJEN"] = stations_actual_reversed["GILZE-RIJEN"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0


class WeatherSourceBase(object):  # pragma: no cover
    """Base class that contains the basic functionality for all sources. Any new sources should implement this as their
    base class!
    """

    def setup_models(self, model_instances):
        self._models = {model.id: model for model in model_instances if not model.async_model}
        self._async_models = {model.id: model for model in model_instances if model.async_model}

    def get_model(self, model_id, fetch_async=False):
        if fetch_async:
            return self._async_models.get(model_id, None)
        return self._models.get(model_id, None)

    def get_models(self, fetch_async=False):
        if fetch_async:
            return list(self._async_models.values())
        return list(self._models.values())

    @property
    def models(self):
        return self.get_models(fetch_async=False)

    @property
    def async_models(self):
        return self.get_models(fetch_async=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Main Application."""

from fastapi import FastAPI
from contextlib import asynccontextmanager
from starlette.responses import RedirectResponse

from weather_provider_api.app_version import APP_VERSION
from weather_provider_api.config import APP_CONFIG
from weather_provider_api.core.initializers.cors import initialize_cors_middleware
from weather_provider_api.core.initializers.exception_handling import (
    initialize_exception_handler,
)
from weather_provider_api.core.initializers.headers import initialize_header_metadata
from weather_provider_api.core.initializers.logging_handler import initialize_logging
from weather_provider_api.core.initializers.mounting import mount_api_version
from weather_provider_api.core.initializers.prometheus import (
    initialize_prometheus_interface,
)
from weather_provider_api.core.initializers.validation import initialize_api_validation
from weather_provider_api.versions.v1 import app as v1
from weather_provider_api.versions.v2 import app as v2


def _build_api_application() -> FastAPI:
    """The main method for building a standardized FastAPI application for use with for instance Uvicorn.

    Returns:
        FastAPI:    A FastAPI application with the full API hooked up and properly initialized.

    Notes:
        This method can also serve as an example for setting up your own version of this API with different settings.

    """
    app_title = APP_CONFIG["base"]["title"]
    app_description = APP_CONFIG["base"]["description"]


    # Define lifespan context manager for startup/shutdown events
    @asynccontextmanager
    async def lifespan(app: FastAPI):
        initialize_logging()
        yield

    # Setting up the base application with lifespan
    application = FastAPI(
        version=APP_VERSION,
        title=app_title,
        summary=app_description,
        description=app_description,
        contact={
            "name": APP_CONFIG["maintainer"]["name"],
            "email": APP_CONFIG["maintainer"]["email_address"],
        },
        lifespan=lifespan,
    )
    application.openapi_version = "3.0.2"

    # Attach selected middleware
    initialize_exception_handler(application)
    initialize_header_metadata(application)
    if APP_CONFIG["base"]["expiration_checking"]:
        initialize_api_validation(application)
    if APP_CONFIG["components"]["prometheus"]:
        initialize_prometheus_interface(application)
    if APP_CONFIG["components"]["cors"]:
        initialize_cors_middleware(application)

    # Create and connect to end-points
    mount_api_version(application, v1)
    mount_api_version(application, v2)

    # Adding a redirect from the root of the application to our default view
    @application.get("/")
    def redirect_to_docs():
        """This function redirects the visitors to the default view from the application's base URL."""
        return RedirectResponse(url="/api/v2/docs")

    application.openapi()

    return application


# Application constant available for re-use, rather than re-initialization.
WPLA_APPLICATION = _build_api_application()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0
from weather_provider_api.routers.weather.sources.knmi.client.actuele_waarnemingen_register_repository import (
    ActueleWaarnemingenRegisterRepository,
)


def main():
    waarnemingen_repo = ActueleWaarnemingenRegisterRepository()
    waarnemingen_repo.update()


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Dict

from fastapi import FastAPI
from gunicorn.app.base import Application
from loguru import logger


class GunicornApplication(Application):
    """Gunicorn Application class.

    This class extends the base Gunicorn application class with the settings and methods needed to directly initialize
     the API application.

    Notes:
          This class can also serve as an example for deploying your own version of this API via Gunicorn.

    """

    def __init__(self, fastapi_application: FastAPI, options: Dict):
        """Overwrite of the base method with the purpose of automatically loading a FastAPI application and deployment
         options passed.

        Args:
            fastapi_application:    The FastAPI application to deploy
            options:                The deployment options to use
        """
        super().__init__()
        self.usage = None
        self.callable = None
        self.options = options
        self.do_load_config()  # Loads the default config and requests extra config settings via the init function
        self.fastapi_application = fastapi_application
        logger.info("Gunicorn application initialized successfully...")

    def init(self, *args):
        """Method overwrite of the base method.

        If effectively loaded upon configuration loading. The returned dictionary holds configuration settings which
         are used to overwrite and/or append self.cfg settings.

        Args:
            *args:  Not actually used.

        Returns:
            (dict): A dictionary holding the configuration settings overwrite or append in self.cfg

        """
        config = {}
        for key, value in self.options.items():
            # Match list to actual settings in self.cfg and add those validated with proper capitalization to the
            #  return value
            if key.lower() in self.cfg.settings and value is not None:
                config[key.lower()] = value

        return config

    def load(self):
        """Overwrite of the base method.

        Used to load the application for use.

        Returns:
            FastAPI:    The FastAPI application to load.

        """
        return self.fastapi_application



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import math
from enum import Enum

"""
Modules translates RC coordinates to WGS84 and vice versa.
This module uses an approximation system originating from a document known as "Transformatieformules.pdf" from Dutch
site http://dekoepel.nl, though the file can no longer be found there.

To convert RC to WGS84 this module uses Function 6 from the original document in combination with Table 4.
To convert WGS84 to RC it uses Function 7 in combination with Table 5.

Because RC doesn't have its center at zero coordinates, we also use Table 3 to properly offset the coordinates for
calculation.
"""


# Coordinate Enum to allow for easy adding of more coordinate types
class CoordinateSystem(str, Enum):
    rd = "RD"
    wgs84 = "WGS84"


class GeoPosition:
    # Zero point configuration (Table 3) based on the center point of the RD coordinate system
    # (phi and lambda shift (very) slightly over the years...)
    X0 = 155000.0
    Y0 = 463000.0
    PHI0 = 52.15517440
    LAMBDA0 = 5.38720621

    # Tables 4 & 5
    K = [
        [0, 1, 3235.65389],
        [2, 0, -32.58297],
        [0, 2, -0.24750],
        [2, 1, -0.84978],
        [0, 3, -0.06550],
        [2, 2, -0.01709],
        [1, 0, -0.00738],
        [4, 0, 0.00530],
        [2, 3, -0.00039],
        [4, 1, 0.00033],
        [1, 1, -0.00012],
    ]
    L = [
        [1, 0, 5260.52916],
        [1, 1, 105.94684],
        [1, 2, 2.45656],
        [3, 0, -0.81885],
        [1, 3, 0.05594],
        [3, 1, -0.05607],
        [0, 1, 0.01199],
        [3, 2, -0.00256],
        [1, 4, 0.00128],
        [0, 2, 0.00022],
        [2, 0, -0.00022],
        [5, 0, 0.00026],
    ]
    R = [
        [0, 1, 190094.945],
        [1, 1, -11832.228],
        [2, 1, -114.221],
        [0, 3, -32.391],
        [1, 0, -0.705],
        [3, 1, -2.340],
        [1, 3, -0.608],
        [0, 2, -0.008],
        [2, 3, 0.148],
    ]
    S = [
        [1, 0, 309056.544],
        [0, 2, 3638.893],
        [2, 0, 73.077],
        [1, 2, -157.984],
        [3, 0, 59.788],
        [0, 1, 0.433],
        [2, 2, -6.439],
        [1, 1, -0.032],
        [0, 4, 0.092],
        [1, 4, -0.054],
    ]

    def __init__(self, xcoord, ycoord, locformat=None):
        self.x = xcoord
        self.y = ycoord
        if locformat is None:
            self.system = self._determine_coordinatesystem()
        elif locformat not in [item for item in CoordinateSystem]:
            self.system = self._determine_coordinatesystem()
        else:
            self.system = locformat

        if self.system == -1:
            raise ValueError("No valid coordinate system could be determined from the coordinates given..")
        if self._out_of_bounds():
            raise ValueError("Invalid coordinates for type were used")

    def _determine_coordinatesystem(self):
        # Checks for each known system whether the coordinates are within a unique range for that system
        # Upon success, it sets the found system
        if (7000 <= self.x <= 300000) and (289000 <= self.y <= 629000):
            return CoordinateSystem.rd
        if (-180 <= self.x <= 180) and (-90 <= self.y <= 90):
            return CoordinateSystem.wgs84

        return -1

    def _out_of_bounds(self):
        # Checks for the set system whether the coordinates are within bound.
        if self.system == CoordinateSystem.wgs84 and (-180 <= self.x <= 180) and (-90 <= self.y <= 90):
            return False
        if (
            self.system == CoordinateSystem.rd
            and (7000 <= self.x <= 300000)
            and (289000 <= self.y <= 629000)
            and (self.x < self.y)
        ):
            return False
        return True

    def get_RD(self):
        # If the system isn't the same as called, return the proper conversion result.
        # Otherwise, return the original values
        if self.system == CoordinateSystem.wgs84:
            return self._wgs84_to_rd()
        return self.x, self.y

    def get_WGS84(self):
        # If the system isn't the same as called, return the proper conversion result.
        # Otherwise, return the original values

        if self.system == CoordinateSystem.rd:
            return self._rd_to_wgs84()
        return self.x, self.y

    def get_original(self):
        return self.x, self.y

    def _wgs84_to_rd(self):
        # Convert WGS84 to RD, using function 7 in combination with the R and S conversion sets.
        dPhi = 0.36 * (self.x - self.PHI0)
        dLambda = 0.36 * (self.y - self.LAMBDA0)
        x = 0
        y = 0

        for row in self.R:
            x = x + row[2] * math.pow(dPhi, row[0]) * math.pow(dLambda, row[1])
        for row in self.S:
            y = y + row[2] * math.pow(dPhi, row[0]) * math.pow(dLambda, row[1])

        x = self.X0 + round(x, 0)
        y = self.Y0 + round(y, 0)

        return x, y

    def _rd_to_wgs84(self):
        # Convert RD to WGS84, using function 6 in combination with the K and L conversion sets.
        dX = (self.x - self.X0) * 0.00001
        dY = (self.y - self.Y0) * 0.00001
        phi = 0.0
        _lambda = 0.0

        for row in self.K:
            phi = phi + row[2] * math.pow(dX, row[0]) * math.pow(dY, row[1])
        for row in self.L:
            _lambda = _lambda + row[2] * math.pow(dX, row[0]) * math.pow(dY, row[1])

        phi = self.PHI0 + (phi / 3600)
        _lambda = self.LAMBDA0 + (_lambda / 3600)

        return phi, _lambda



#!/usr/bin/env python


#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0
import sys

from loguru import logger

from weather_provider_api.routers.weather.sources.cds.client.era5sl_repository import (
    ERA5SLRepository,
)


def main(args) -> None:
    """Run the update of the ERA5SL repository."""
    test_mode = False

    if len(args) == 2 and args[1] == "testmode":
        logger.warning("WARNING: Running in test mode")
        test_mode = True

    era5sl_repo = ERA5SLRepository()
    era5sl_repo.update(test_mode)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
import glob
from datetime import datetime

from dateutil.relativedelta import relativedelta
from loguru import logger
from pytz import UTC

from weather_provider_api.routers.weather.repository.repository import RepositoryUpdateResult, WeatherRepositoryBase
from weather_provider_api.routers.weather.sources.cds.client.cds_api_tools import CDSDataSets
from weather_provider_api.routers.weather.sources.cds.client.era5_utils import (
    Era5UpdateSettings,
    era5_repository_update,
)
from weather_provider_api.routers.weather.sources.cds.factors import era5sl_factors
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.grid_helpers import round_coordinates_to_wgs84_grid


class ERA5SLRepository(WeatherRepositoryBase):
    """A class that holds all functionality (excepting the downloader) for the ERA5 Single Levels Repository."""

    def __init__(self):
        """Initializes the ERA5 Single Levels Repository."""
        super().__init__()
        self.repository_name = "CSD ERA5 Single Levels"
        self.file_prefix = "ERA5SL"
        self.runtime_limit = 3 * 60  # 3 hours maximum runtime
        self.permanent_suffixes = ["INCOMPLETE", "TEMP"]
        self.grid_resolution = 0.25
        self.file_identifier_length = 7
        self.age_of_permanence_in_months = 3
        self.age_of_permanence_in_months = 3
        logger.debug(f"Initialized {self.repository_name} repository")

    @staticmethod
    def _get_repo_sub_folder() -> str:
        """Returns the subfolder name for the repository."""
        return "ERA5_SL"

    @property
    def first_day_of_repo(self) -> datetime:
        """Returns the first day of the repository."""
        first_day_of_repo = datetime.now(UTC) - relativedelta(years=12, days=5)
        first_day_of_repo = first_day_of_repo.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        return first_day_of_repo

    @property
    def last_day_of_repo(self) -> datetime:
        """Returns the last day of the repository."""
        last_day_of_repo = datetime.now(UTC) - relativedelta(days=5)
        last_day_of_repo = last_day_of_repo.replace(hour=0, minute=0, second=0, microsecond=0)
        return last_day_of_repo

    def update(self, test_mode: bool) -> RepositoryUpdateResult:
        """The update implementation for the ERA5 Single Levels repository.

        This function handles all the required actions to update the repository completely, but taking into
        account its set runtime_limit. If based on the time of completion of other downloaded files this session
        the next file wouldn't complete within the runtime_limit, the update process halts.
        (if no other downloads were made yet, a generous rough estimate is used).

        Args:
            test_mode: A boolean indicating whether the update process should run in test mode.

        Returns:
            A RepositoryUpdateResult value indicating a completion, time-out or failure of the update process
        """
        # Always start with a nicely cleaned repository
        self.cleanup()

        logger.info(f"ERA5 Single Levels Update - Storage in: {self.repository_folder} ")
        return era5_repository_update(
            Era5UpdateSettings(
                filename_prefix=self.file_prefix,
                era5_dataset_to_update_from=CDSDataSets.ERA5SL,
                era5_product_type="reanalysis",
                factor_dictionary=era5sl_factors,
                factors_to_process=[era5sl_factors[x] for x in list(era5sl_factors.keys())],
                maximum_runtime_in_minutes=self.runtime_limit,
                repository_time_range=(self.first_day_of_repo, self.last_day_of_repo),
                target_storage_location=self.repository_folder,
            ),
            test_mode=test_mode,
        )

    def _delete_files_outside_of_scope(self):
        """A function that deletes all files in the repository with a date not inside the repository's scope.

        All files labeled as either before or after the given scope will be deleted.

        Returns:
            Nothing. Successful means the all files outside the scope were deleted.
        """
        len_filename_until_date = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1

        for file_name in glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.nc"):
            file_year = int(file_name[len_filename_until_date : len_filename_until_date + 4])
            file_month = int(file_name[len_filename_until_date + 5 : len_filename_until_date + 7])

            if (
                file_year < self.first_day_of_repo.year
                or file_year > self.last_day_of_repo.year
                or (file_year == self.first_day_of_repo.year and file_month < self.first_day_of_repo.month)
                or (file_year == self.last_day_of_repo.year and file_month > self.last_day_of_repo.month)
            ):
                logger.debug(
                    f"Deleting file [{file_name}] because it does not lie in the "
                    f"repository scope ({self.first_day_of_repo, self.last_day_of_repo})"
                )
                self._safely_delete_file(file_name)

    def _get_file_list_for_period(self, start: datetime, end: datetime):
        """A function that retrieves a list of files in the repository associated with the requested period of time.

        Args:
            start:  A datetime containing the start of the requested period of time.
            end:    A datetime containing the end of the requested period of time.

        Returns:
            A list of files (in string format) that indicate the files containing data for the requested period.
        """
        self.cleanup()

        len_filename_until_date = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1
        logger.info(
            f"Searching for ERA5 Single Levels files in repository folder: "
            f"{self.repository_folder.joinpath(self.file_prefix)}"
        )
        full_list_of_files = glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.nc")
        list_of_filtered_files = []
        for file in full_list_of_files:
            file_year = int(file[len_filename_until_date : len_filename_until_date + 4])
            file_month = int(file[len_filename_until_date + 5 : len_filename_until_date + 7])
            date_for_filename = datetime(year=file_year, month=file_month, day=15).astimezone(UTC)

            if (
                start.replace(day=1)
                < date_for_filename
                < datetime(year=end.year, month=end.month, day=28).astimezone(UTC)
            ):
                # If the file is within the requested period, save it to the list of filtered files
                list_of_filtered_files.append(file)

        return list_of_filtered_files

    def get_grid_coordinates(self, coordinates: list[GeoPosition]) -> list[GeoPosition]:
        """Round a list of GeoPositions to the resolution set through grid_resolution."""
        return round_coordinates_to_wgs84_grid(
            coordinates=coordinates,
            grid_resolution_lat_lon=(self.grid_resolution, self.grid_resolution),
            starting_points_lat_lon=(50.75, 3.2),  # Used to properly round values
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import glob
import shutil
from abc import ABCMeta, abstractmethod
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import List

import xarray as xr
from fastapi import HTTPException
from loguru import logger

from weather_provider_api.config import APP_STORAGE_FOLDER
from weather_provider_api.core.initializers.exception_handling import (
    NOT_IMPLEMENTED_ERROR,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class RepositoryUpdateResult(Enum):
    failure = 0
    completed = 1
    timed_out = 2


class WeatherRepositoryBase(metaclass=ABCMeta):
    """This is the base class for all weather data storage repositories. Any new repositories should implement this
    as their base class.
    All valid stored repository files are named as follows:
    {file_prefix}_{file_identifier}.nc
    or
    {file_prefix}_{file_identifier}_{permanent_suffix}.nc
    Any files found not matching this pattern shall be deleted as being temporary in nature.
    """

    def __init__(self):
        """Specification of required fields for a weather data storage repository:
        - repository_folder:    Contains the folder where the repository will be saved. Is set inside a main
                                repository folder, and based on a sub-folder passed by the repository itself
                                through the _get_repo_subfolder() function.
        - file_prefix:          Contains a string with the file_prefix to use for all files within the
                                repository. Is set from the repository itself.
        - runtime_limit:        Contains the maximum time the update function of the repository is allowed to be
                                running, in seconds.
        - first_day_of_repo:    Contains a datetime indicating the oldest moment allowed to be stored in the
                                repository.
        - last_day_of_repo:     Contains a datetime indicating the newest moment allowed to be stored in the
                                repository.
        - permanent_suffixes:   Contains a list of suffixes that can be added to the repository files to
                                indicate that the file should not be deleted. Any file not matching the prefix
                                or having a suffix not matching this list will be deleted upon cleanup.
        - file_identifier_length:   This is the length in characters that the unique identifier part of the
                                    filename takes up. Usually this is based on a datetime.
        """
        self.repository_folder = Path(APP_STORAGE_FOLDER).joinpath(self._get_repo_sub_folder())
        self.repository_name = None
        self.file_prefix = None
        self.runtime_limit = 60 * 60 * 2  # seconds * minutes * hours (2 hours default)
        self.permanent_suffixes = None
        self.file_identifier_length = None

    @property
    def first_day_of_repo(self):
        raise NotImplementedError()

    @property
    def last_day_of_repo(self):
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def _get_repo_sub_folder():
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    def _validate_repo_folder(self):
        """This function checks whether the repository folder already exists (starting from its parent folder)
        and creates it, if it (or its parent folder) don't exist yet.
        """
        if not Path(self.repository_folder).exists():  # If the folder doesn't exist yet, create it
            logger.debug(f"Attempting to create folder[{self.repository_folder}]")
            try:
                # If the main folder for all repositories doesn't exist yet, create it
                if not Path(self.repository_folder).parent.exists():
                    Path(self.repository_folder).parent.mkdir()
                Path(self.repository_folder).mkdir()
            except OSError as e:
                logger.error(f"An error occurred creating the directory: {e}")
                raise e

    def cleanup(self):
        """This is the cleanup function for any Weather Repository.
        Any files not matching the pattern required for the Repository shall be deleted.
        """
        self._validate_repo_folder()
        logger.debug(f"Verifying existing files for {self.repository_name} in [{self.repository_folder}]")

        # Delete any files that aren't of a permanent type
        self._delete_non_permanent_files()

        # First we delete any files that are too old or new to be valid
        self._delete_files_outside_of_scope()

        # Only one file may exist per month. Select the proper file to remain and remove any others
        self._delete_excess_files()

    @abstractmethod
    def update(self, test_mode: bool = False) -> RepositoryUpdateResult:
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    def gather_period(self, begin: datetime, end: datetime, coordinates: List[GeoPosition]) -> xr.Dataset:
        """A function that gathers the repository files associated with a requested period, and then returns the full
            weather data that matches both that period as the requested locations from those files, as a Xarray Dataset
        Args:
            begin:          A datetime holding the starting moment for the requested period to gather data for
            end:            A datetime holding the ending moment for the requested period to gather data for
            coordinates:    A list of GeoPositions holding the coordinates that the data request is for
        Returns:
            A Xarray Dataset containing all the repository data that matches both the requested period, and
            the requested coordinates.
        """
        self.cleanup()
        logger.debug(f"Gathering repository data for the period of {begin} to {end}")

        # Get a list of files matching the requested period
        file_list = self._get_file_list_for_period(begin, end)

        if len(file_list) == 0:
            logger.error(f"No files were found for the period of {begin} to {end}")
            raise HTTPException(
                404,
                f"No data was found for the period of [{begin.date()}] to [{end.date()}] in "
                f"repository [{self.repository_name}]. As this range lies within the repository's "
                f"storage timeframe of [{self.first_day_of_repo}] to [{self.last_day_of_repo}], "
                f"we suggest retrying in a couple of days. If still no data is found at that time,"
                f"please contact us at [weather.provider@alliander.com].",
            )

        # Load files into datasets, select the requested data and aggregate that into a single dataset
        ds = xr.Dataset()
        for file in file_list:
            logger.debug(f"Processing file: {file}")
            ds_temp = xr.open_dataset(file).load()

            ds_temp = self._filter_dataset_by_coordinates(coordinates, ds_temp)
            if file == file_list[0]:
                ds = ds_temp
            else:
                ds = ds.combine_first(ds_temp)
        return ds

    @staticmethod
    def load_file(file: Path) -> xr.Dataset:
        """A function that loads and returns the full data for a specific repository file as a Xarray Dataset
        Args:
            file:   The filename (in the Path format by PathLib) specifying the file to load
        Returns:
            An Xarray Dataset containing all the weather data held within the specified file.
        """
        if file.exists():
            with xr.open_dataset(file) as ds:
                ds.load()
            return ds

        # Raise a FileNotFoundError if the file doesn't exist
        logger.error(f"File [{file!s} does not exist]")
        raise FileNotFoundError

    def purge_repository(self):
        """Function to fully delete the repository's folder and create a new clean one. Use with care!"""
        logger.warning(f"Purging the entire repository folder for {self.repository_name}!")
        shutil.rmtree(self.repository_folder, ignore_errors=True)
        self._validate_repo_folder()  # Rebuild the folder after deletion

    def _delete_non_permanent_files(self):
        """A function that deletes any and all files in the repository's folder that are not considered permanent in
        nature. Only the files matching either repository files without a suffix or those with suffix listed in the
        permanent_suffixes field are allowed.
        Every other file should be deleted from the repository immediately.
        """
        # TODO: Enhance the detection of files that do not belong in the folder.
        #       A repository folder should only have repository files..
        len_filename_until_after_date = (
            len(str(self.repository_folder.joinpath(self.file_prefix))) + self.file_identifier_length + 2
        )
        for file_name in glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.nc"):
            file_suffix = file_name[len_filename_until_after_date:-3]
            if len(file_suffix) != 0 and file_suffix not in self.permanent_suffixes:
                logger.debug(
                    f"File [{file_name}] is not a permanent file for {self.repository_name} and needs to be deleted"
                )
                self._safely_delete_file(file_name)

    @abstractmethod
    def _delete_files_outside_of_scope(self):
        pass

    def _delete_excess_files(self):
        """A function that selects the proper file to keep when more than one permanent file exists for a given
        identifier. The other files are deleted.
        """
        len_filename_until_date = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1
        file_list = glob.glob(str(self.repository_folder.joinpath(self.file_prefix)) + "*.*")
        identifier_list = list(
            set(
                [
                    file[len_filename_until_date : len_filename_until_date + self.file_identifier_length]
                    for file in file_list
                ]
            )
        )

        for identifier in identifier_list:
            files_with_specific_identifier = glob.glob(
                str(self.repository_folder.joinpath(self.file_prefix)) + "_" + identifier + "*.nc"
            )

            if len(files_with_specific_identifier) > 1:
                logger.debug(f"More than one file was found for identifier [{identifier}]")
                file_to_retain = None
                highest_ranking_suffix = None

                for file in files_with_specific_identifier:
                    suffix = file[len_filename_until_date + self.file_identifier_length + 1 : -3]

                    if highest_ranking_suffix is None or suffix == "":
                        highest_ranking_suffix = suffix
                        file_to_retain = file

                    # TEMP trumps INCOMPLETE because in the normal process incomplete files will always be replaced with
                    # temporary files
                    if highest_ranking_suffix == "INCOMPLETE" and suffix == "TEMP":
                        highest_ranking_suffix = suffix
                        file_to_retain = file

                for file in files_with_specific_identifier:
                    if file != file_to_retain:
                        self._safely_delete_file(str(Path(file)))

    @staticmethod
    def _safely_delete_file(file: str):
        """Basic function to safely remove files from the repository if possible, and supply errors if not"""
        try:
            logger.debug(f"Safely deleting file [{file}]")
            Path(file).unlink()
        except OSError as e:
            logger.error(f"Could not safely delete file: {e}")
            raise OSError(f"Could not safely delete file: {file}")
        return True

    @abstractmethod
    def _get_file_list_for_period(self, start: datetime, end: datetime):
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)

    def _filter_dataset_by_coordinates(self, coordinates: List[GeoPosition], ds: xr.Dataset) -> xr.Dataset:
        """A function that filters a given Xarray Dataset down to the values matching a given list of locations.

        Args:
            coordinates:    A list of GeoPositions that the data is requested for.
            ds:             An Xarray Dataset containing the data to be filtered.

        Returns:
            An Xarray Dataset containing only the weather data that matched the given list of coordinates.
        """
        ds_selected = xr.Dataset()
        coordinate_list = self.get_grid_coordinates(coordinates)

        for coordinate in coordinate_list:
            # First filter a single coordinate in the list
            ds_single_coord = ds.stack(dimensions={"coord": ["lat", "lon"]})
            ds_single_coord = ds_single_coord.where(ds_single_coord.lat == coordinate.get_WGS84()[0], drop=True)
            ds_single_coord = ds_single_coord.where(
                ds_single_coord.lon.round(3) == coordinate.get_WGS84()[1].round(3),
                drop=True,
            )
            ds_single_coord = ds_single_coord.unstack("coord")
            # Then append this to a clean list
            if coordinate == coordinate_list[0]:
                ds_selected = ds_single_coord
            else:
                ds_selected = ds_selected.combine_first(ds_single_coord)

        return ds_selected

    @abstractmethod
    def get_grid_coordinates(self, coordinates: List[GeoPosition]) -> List[GeoPosition]:
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""CDS - ERA5 Single Levels Weather data Model"""

import copy
from datetime import datetime
from typing import List, Optional

import numpy as np
import xarray as xr
from loguru import logger

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.cds.client.era5sl_repository import (
    ERA5SLRepository,
)
from weather_provider_api.routers.weather.sources.cds.factors import era5sl_factors
from weather_provider_api.routers.weather.utils.date_helpers import (
    validate_begin_and_end,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class ERA5SLModel(WeatherModelBase):
    """A Weather Model that incorporates the:
        ERA5 hourly data on single levels from 1979 to present
    dataset into the Weather Provider API
    """

    def __init__(self):
        super().__init__()
        self.id = "era5sl"
        logger.debug(f"Initializing weather model [{self.id}]")

        self.name = "CDS ERA5 - Hourly data on single levels from 1979 to the present"

        self.version = "0.3"
        self.url = "https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview"
        self.predictive = False
        self.description = (
            "Hourly weather measurements. Can be returned for a specified period. The number of "
            "measurements returned depends on this period selection."
        )
        self.async_model = False

        self.time_step_size_minutes = 60
        self.num_time_steps = 0

        self.repository = ERA5SLRepository()

        # Set up Conversion Dictionary
        si_conversion_dict = {
            k: {"name": k, "convert": lambda x: x} for k in era5sl_factors.values()
        }  # The default output format for ERA5SL is already SI
        self.to_si = si_conversion_dict

        # Human output conversion:
        self.to_human = copy.deepcopy(self.to_si)
        # self.to_human["sea_surface_temperature"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_1"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_2"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_3"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_4"]["convert"] = self.kelvin_to_celsius
        self.to_human["2m_temperature"]["convert"] = self.kelvin_to_celsius

        logger.debug(f"Weather model [{self.id}] initialized successfully")

    def is_async(self):
        """Returns the async model status"""
        return self.async_model

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: datetime,
        end: datetime,
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested ERA5 Single Levels weather data from the repository
            and returns it as a Xarray Dataset.

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.
        """
        # Test and account for invalid datetime timeframes or input
        begin, end = validate_begin_and_end(
            begin,
            end,
            self.repository.first_day_of_repo,
            self.repository.last_day_of_repo,
        )

        # Validate the requested weather factors:
        validated_factors = self._validate_weather_factors(weather_factors)

        ds = self._fill_dataset_with_data(coords, begin, end, validated_factors)
        return ds

    @staticmethod
    def _validate_weather_factors(weather_factors: List[str]) -> List[str]:
        """A function that validates a list of weather factors to that of the dataset in the repository.
            Existing factors will be kept, non-existing removed, and if the list is empty the full set for the dataset
            will be used.

        Args:
            weather_factors:    A list of weather factors to validate (in string format)

        Returns:
            A list of weather factors (in string format) only factors that match those of the ERA5SL dataset.
        """
        if weather_factors is None:
            weather_factors = [era5sl_factors[x] for x in list(era5sl_factors.keys())]

        # Lookup using the generic long name
        weather_factors_long_names = [x for x in weather_factors if x in era5sl_factors.values()]
        # Lookup using the CDS' own short name
        weather_factors_short_names = [era5sl_factors[x] for x in weather_factors if x in era5sl_factors.keys()]

        # Merge the results
        weather_factors = weather_factors_long_names + weather_factors_short_names

        # If nothing useful was found, just return everything
        return weather_factors

    @staticmethod
    def _get_list_of_factors_to_drop(factors: List[str]) -> List[str]:
        # A small function that that compares a list of factors to keep with the full list, to make a list of factors
        # to drop from a full set.
        to_drop = [x for x in era5sl_factors.values() if x not in factors]
        logger.debug("Dropping the following factors for the request: " + str(to_drop))
        return to_drop

    def _fill_dataset_with_data(
        self,
        era5sl_coordinates: List[GeoPosition],
        begin: datetime,
        end: datetime,
        validated_factors: List[str],
    ) -> xr.Dataset:
        """A function that fills a dataset with ERA5SL weather data from the repository, based on the requested
            coordinates and period, and removes any not-requested weather factors from the output.

        Args:
            era5sl_coordinates:     A list of GeoPositions containing the locations to be gathered from the repository.
            begin:                  A datetime containing the starting moment for the period to gather.
            end:                    A datetime containing the end moment for the period to gather.
            validated_factors:      A list of valid ERA5SL factors wanted for the result.

        Returns:
            An Xarray Dataset containing the weather data requested.
        """
        # Gather a dataset with the proper period and coordinates
        ds = self.repository.gather_period(begin, end, era5sl_coordinates)

        ds = ds.sel(time=slice(np.datetime64(begin), np.datetime64(end)))

        # Drop excess weather factors
        ds = ds.drop_vars(self._get_list_of_factors_to_drop(validated_factors))
        return ds

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()

            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Base Model class.

An expansion of the PydanticBaseModel used to build each Weather Provider Base class.
"""

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    """Minimal PydanticBaseMode expansion used in all base classes."""

    class Config:
        from_attributes = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.routers.weather.sources.knmi.client.arome_repository import (
    HarmonieAromeRepository,
)


def main():
    arome_repo = HarmonieAromeRepository()
    arome_repo.update()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

""" """

from datetime import datetime
from typing import List

import accept_types
from fastapi import APIRouter, BackgroundTasks, Depends, Header, HTTPException, Request
from loguru import logger

from weather_provider_api.core.initializers.rate_limiter import API_RATE_LIMITER
from weather_provider_api.routers.weather.api_models import (
    WeatherContentRequestMultiLocationQuery,
    WeatherContentRequestQuery,
    WeatherFormattingRequestQuery,
    WeatherModel,
    WeatherSource,
    result_mime_types,
)
from weather_provider_api.routers.weather.controller import WeatherController
from weather_provider_api.routers.weather.sources.weather_alert.weather_alert import (
    WeatherAlert,
)
from weather_provider_api.routers.weather.utils import serializers
from weather_provider_api.routers.weather.utils.date_helpers import parse_datetime
from weather_provider_api.routers.weather.utils.file_helpers import remove_file

v2_router = APIRouter()

controller = WeatherController()


def header_accept_type(accept: str = Header(None)) -> str:
    # Matches the given string to the nearest result in result_mime_types and returns the matching value
    best_match_or_none = accept_types.get_best_match(accept, result_mime_types)
    return result_mime_types[best_match_or_none]


# @weather_provider_api.get("/sources", response_model=List[WeatherSource], tags=["sync", "async"])
@v2_router.get("/sources", response_model=List[WeatherSource], tags=["sync"])
async def get_sources():  # pragma: no cover
    """<B>List all the Weather Sources available</B>"""
    """
        An API function that returns all of the Sources available from the WeatherController
    Args:
        None
    Returns:
        A list of all Source available from the WeatherController
    """
    return controller.get_sources()


# @weather_provider_api.get("/sources/{source_id}", response_model=WeatherSource, tags=["sync", "async"])
@v2_router.get("/sources/{source_id}", response_model=WeatherSource, tags=["sync"])
async def get_source(source_id: str):  # pragma: no cover
    """<B>List all the Models available  for the Source</B>"""
    """
        An API function that returns all of the models available for the given source  
    Args:
        source_id: The source ID for the given source
    Returns:
        A list of all Models available for the given source
    """
    source_id = source_id.lower()
    return controller.get_source(source_id)


@v2_router.get("/sources/{source_id}/models", response_model=List[WeatherModel], tags=["sync"])
async def get_sync_models(source_id: str):  # pragma: no cover
    """<B>List all the synchronous Models available for the selected Source</B>"""
    """
        An API function that returns all of the synchronous models available for the given source
    Args:
        source_id: The source ID for the given source
    Returns:
        A list of all synchronous Models available for the given source
    """

    source_id = source_id.lower()
    return controller.get_models(source_id, fetch_async=False)


@v2_router.get("/sources/{source_id}/models/{model_id}", tags=["sync"])
@API_RATE_LIMITER.limit("20/minute")
async def get_sync_weather(
    request: Request,
    source_id: str,
    model_id: str,
    cleanup_tasks: BackgroundTasks,
    ret_args: WeatherContentRequestQuery = Depends(),
    fmt_args: WeatherFormattingRequestQuery = Depends(),
    accept: str = Depends(header_accept_type),
):  # pragma: no cover
    """<B>Request weather data for a specific Model using the given settings (location, period, weather factors, e.g.). <BR>
    This data is then formatted as the requested output format (output unit system and file format) before returning the requested data.</B>

    <I>(Please note that as some models are predictive or otherwise restricted in the periods available for requests,
    that sometimes the 'begin' and 'end' values will be altered to match these restrictions.)</I>
    """
    """
        An API function that retrieves specific weather data for a specific Weather Model and returns it as the
        requested output format and units.
    Args:
        source_id:      The Source ID of the Source to request the weather data from.
        model_id:       The Model ID for the Model to request the weather data from.
        cleanup_tasks:  A BackgroundTasks object to hold any pending cleanup tasks for when the data request is
                        finished.
        ret_args:       A WeatherContentRequestQuery object holding the parameters for the weather data request to
                        use.
        fmt_args:       A WeatherFormattingRequestQuery object holding the parameters for the output format and
                        units to use.
        accept:         Header type to use for the output file.
                        Rounded to the most likely value using header_accept_type().
    Returns:
        The weather data in the requested format for the requested parameters.
    """
    starting_time = datetime.utcnow()
    logger.info(f"WeatherRequest({starting_time}): {request.url}")
    source_id = source_id.lower()
    model_id = model_id.lower()
    coords = controller.lat_lon_to_coords(ret_args.lat, ret_args.lon)

    begin = parse_datetime(ret_args.begin, raise_errors=True, loc=["query", "begin"])
    end = parse_datetime(
        ret_args.end,
        round_missing_time_up=True,
        raise_errors=True,
        loc=["query", "end"],
    )

    try:
        weather_data = controller.get_weather(
            source_id,
            model_id,
            fetch_async=False,
            coords=coords,
            begin=begin,
            end=end,
            factors=ret_args.factors,
        )
    except FileNotFoundError as e:
        raise HTTPException(status_code=404, detail=e.args[0])

    if weather_data is None:
        raise HTTPException(status_code=404, detail="No data was found for the given period")

    response_format = fmt_args.response_format or accept

    converted_weather_data = controller.convert_names_and_units(
        source_id, model_id, False, weather_data, fmt_args.units
    )

    coords = [
        (lat_val, lon_val)
        for (lat_val, lon_val) in zip(
            converted_weather_data.coords["lat"].values,
            converted_weather_data.coords["lon"].values,
        )
    ]

    response, optional_file_path = serializers.file_or_text_response(
        converted_weather_data, response_format, source_id, model_id, ret_args, coords
    )
    cleanup_tasks.add_task(remove_file, optional_file_path)
    logger.info(f"WeatherRequest({starting_time}) data preparation finished.")
    return response


# @weather_provider_api.get("/weeralarm")
@v2_router.get("/alarms/knmi", tags=["alerts"])
async def get_alarm():  # pragma: no cover
    """<B>Fetches the WeatherAlarm status for all the provinces from KNMI Weer Alarm and returns the results</B>"""
    """
        An API Function that reads the status of the Weer Alarm pages on the KNMI Site and returns them.
    Returns:
        A formatted text containing the current Weather Alert Status for each of the provinces.
    """
    weather_alert = WeatherAlert()
    return weather_alert.get_alarm()


# Handler for requests with multiple locations:
@v2_router.get("/sources/{source_id}/models/{model_id}/multiple-locations/", tags=["sync"])
@API_RATE_LIMITER.limit("5/minute")
async def get_sync_weather_multi_loc(
    request: Request,
    source_id: str,
    model_id: str,
    cleanup_tasks: BackgroundTasks,
    ret_args: WeatherContentRequestMultiLocationQuery = Depends(),
    fmt_args: WeatherFormattingRequestQuery = Depends(),
    accept: str = Depends(header_accept_type),
):  # pragma: no cover
    starting_time = datetime.utcnow()
    logger.info(f"WeatherRequest({starting_time}): {request.url}")

    source_id = source_id.lower()
    model_id = model_id.lower()

    coords = controller.str_to_coords(ret_args.locations)

    begin = parse_datetime(ret_args.begin, raise_errors=True, loc=["query", "begin"])
    end = parse_datetime(
        ret_args.end,
        round_missing_time_up=True,
        raise_errors=True,
        loc=["query", "end"],
    )

    try:
        weather_data = controller.get_weather(
            source_id,
            model_id,
            fetch_async=False,
            coords=coords,
            begin=begin,
            end=end,
            factors=ret_args.factors,
        )
    except FileNotFoundError as e:
        raise HTTPException(status_code=404, detail=e.args[0])

    if weather_data is None:
        raise HTTPException(status_code=404, detail="No data was found for the given period")

    response_format = fmt_args.response_format or accept

    converted_weather_data = controller.convert_names_and_units(
        source_id, model_id, False, weather_data, fmt_args.units
    )

    new_coords = []
    for coord in coords:
        new_coords.append(coord[0])

    response, optional_file_path = serializers.file_or_text_response(
        converted_weather_data, response_format, source_id, model_id, ret_args, coords
    )
    cleanup_tasks.add_task(remove_file, optional_file_path)
    logger.info(f"WeatherRequest({starting_time}) data preparation finished.")
    return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

# factor name mapping

import json

from weather_provider_api.routers.weather.utils.file_helpers import (
    get_var_map_file_location,
)

file_to_use_era5sl = get_var_map_file_location("era5sl_var_map.json")
file_to_use_era5land = get_var_map_file_location("era5land_var_map.json")

with open(file_to_use_era5sl, "r") as _f:
    era5sl_factors: dict = json.load(_f)
    # for factor in ["mwd", "mwp", "swh"]:
    #     era5sl_factors.pop(factor)


with open(file_to_use_era5land, "r") as _f:
    era5land_factors = json.load(_f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Entry point to the weather provider."""

import datetime
import re
from typing import List, Optional, Tuple

import numpy as np
import xarray as xr

from weather_provider_api.routers.weather.api_models import OutputUnit
from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.base_models.source import WeatherSourceBase
from weather_provider_api.routers.weather.exceptions import (
    UnknownModelException,
    UnknownSourceException,
)
from weather_provider_api.routers.weather.sources.cds.cds import CDS
from weather_provider_api.routers.weather.sources.knmi.knmi import KNMI
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class WeatherController(object):  # pragma: no cover
    """A Controller Object that can handle any request for the hooked up weather models and format it into any of the
    hooked up output formats.
    """

    def __init__(self):
        # Load Sources (and with it their models) and hook them up to the controller
        source_instances = [KNMI(), CDS()]
        self.sources = {source.id: source for source in source_instances}

    def get_weather(
        self,
        source_id: str,
        model_id: str,
        fetch_async: bool,
        coords: list[list[tuple[float, float]]],
        begin: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        factors: list[str] | None = None,
    ):
        """Function to use the requested weather model from the requested source to get specific weather factors for a
            specific time and specific location(s)

        Args:
            source_id:  The weather source that need to be queried (e.g.: knmi, cds)
            model_id:   The model identifier of the model that needs to be queried (has to be a model that exists within
                        the specific source requested through source_id
            fetch_async:    A boolean indicated if the request was made to asynchronously fetch the data or not
            coords:     A nested 3-layer list representing a list of polygons
                        in the case of points, they are treated as a one-point polygon
                        D0: different polygons
                        D1: sequence of points in each polygon
                        D2: coordinates of each point (lat, lon in coordinates)
            begin:      The starting time of the requested output data
            end:        The ending time of the requested output data
            factors:    A list of the requested weather factors for the output (default is all available)

        Returns:
            A Xarray Dataset containing the weather data for the selected model, period(s), location(s) and factor(s)
        """
        model = self.get_model(source_id, model_id, fetch_async)
        coords = self._calculate_polygon_means(coords)  # polygonal areas to mean calculation
        coords = self._tuples_to_geo_positions(coords)  # conversion of Tuples with floats to GeoPosition items

        if model.is_async():
            return None

        # call relevant routine data
        ds = model.get_weather(coords=coords, begin=begin, end=end, weather_factors=factors)
        return ds

    def convert_names_and_units(
        self,
        source_id: str,
        model_id: str,
        fetch_async: bool,
        weather_data: xr.Dataset,
        unit: OutputUnit,
    ):
        # A function that uses the indicated model's built-in conversion method to alter the values in a given dataset
        # to match the requested output unit format.
        model = self.get_model(source_id, model_id, fetch_async)
        ds = model.convert_names_and_units(weather_data, unit)
        return ds

    @staticmethod
    def lat_lon_to_coords(lat: float, lon: float) -> List[List[Tuple[float, float]]]:
        return [[(lat, lon)]]

    @staticmethod
    def str_to_coords(locations_string: str):
        str_coordinates_list = re.findall(r"\(\d{1,3}.?\d*,\s?\d{1,3}.?\d*\)", locations_string)
        coordinate_list = []
        for str_coordinate in str_coordinates_list:
            coordinate = str_coordinate[1:-1].replace(" ", "")
            coordinate = coordinate.split(",")
            coordinate = (float(coordinate[0]), float(coordinate[1]))
            coordinate_list.append([coordinate])

        return coordinate_list

    def get_source_keys(self) -> List[str]:
        return list(self.sources.keys())

    def get_sources(self) -> List[WeatherSourceBase]:
        return list(self.sources.values())

    def get_source(self, source_id: str) -> Optional[WeatherSourceBase]:
        self._validate_source(source_id)
        return self.sources.get(source_id, None)

    def get_models(self, source_id: str, fetch_async: bool = False) -> List[WeatherModelBase]:
        source = self.get_source(source_id)
        return source.get_models(fetch_async)

    def get_model(self, source_id: str, model_id: str, fetch_async: bool = False) -> Optional[WeatherModelBase]:
        self._validate_source_and_model(source_id, model_id, fetch_async)
        source = self.get_source(source_id)
        return source.get_model(model_id, fetch_async)

    def _validate_source(self, source_id: str):
        if source_id not in self.sources:
            raise UnknownSourceException

    def _validate_source_and_model(self, source_id: str, model_id: str, fetch_async: bool = False):
        self._validate_source(source_id)
        if self.sources[source_id].get_model(model_id, fetch_async) is None:
            raise UnknownModelException

    @staticmethod
    def _calculate_polygon_means(
        coords: List[List[Tuple[float, float]]],
    ) -> List[Tuple[float, float]]:
        # Calculating the mean points of polygons containing locations
        coords = [
            (
                float(np.mean([single_point[0] for single_point in single_polygon])),
                float(np.mean([single_point[1] for single_point in single_polygon])),
            )
            for single_polygon in coords
        ]
        return coords

    @staticmethod
    def _tuples_to_geo_positions(
        coords: List[Tuple[float, float]],
    ) -> List[GeoPosition]:
        # Convert the Tuples in a list to a list of Geo Positions
        coords = [GeoPosition(coordinate[0], coordinate[1]) for coordinate in coords]
        return coords
